#ifndef _Function_h_
#define _Function_h_

class Symbol;
class SymbolTable;
class AddressIterator;
class Function;
class FlowGraph;
class BasicBlock;
class Instruction;
class Operand;

#include <Base.h>
#include <BitSet.h>
#include <LinkedList.h>
#include <Instruction.h>
#include <SectHeader.h>
#include <RawSection.h>

class MemoryOperation {
private:
    Instruction instruction;
    uint32_t offset;
    BasicBlock* basicBlock;
    uint32_t index;
    HashCode hashCode;
public:
    MemoryOperation(Instruction insn,uint64_t addr,BasicBlock* bb,uint32_t idx);
    ~MemoryOperation() {}
    bool isDForm();
    bool isDsForm();
    bool isXForm();
    void print();
    Operand getAddressOperand1();
    Operand getAddressOperand2();
    HashCode getHashCode() { return hashCode; }
    BasicBlock* getBasicBlock() { return basicBlock; }
    uint32_t getIndex() { return index; }
    uint64_t getInsnAddr();
    Instruction getInstruction() { return instruction; }
};

class BasicBlock {
private:
    const static uint32_t EntryMask     = 0x1;
    const static uint32_t ExitMask      = 0x2;
    const static uint32_t TraceMask     = 0x4;
    const static uint32_t JumpTableMask = 0x8;
    const static uint32_t NoPathMask    = 0x10;

protected:

    FlowGraph* flowGraph;
    uint32_t index;

    uint64_t baseAddress;
    uint32_t sizeInBytes;

    uint32_t flags;

    HashCode hashCode;

    uint32_t numOfMemoryOps;
    MemoryOperation** memoryOps;
    uint32_t numOfFloatPOps;

    uint32_t numOfSources;
    BasicBlock** sources;
    uint32_t numOfTargets;
    BasicBlock** targets;

    BasicBlock* immDominatedBy;
public:
    BasicBlock(FlowGraph* cfg,uint64_t addr)
            : flowGraph(cfg),index(0),baseAddress(addr),
              sizeInBytes(0),flags(0),
              numOfMemoryOps(0),numOfFloatPOps(0),
              numOfSources(0),sources(NULL),numOfTargets(0),targets(),immDominatedBy(NULL) {}

    ~BasicBlock() {}

    bool inRange(uint64_t addr) { return ((baseAddress <= addr) && (addr < (baseAddress+sizeInBytes))); }
    AddressIterator getInstructionIterator();

    FlowGraph* getFlowGraph() { return flowGraph; }
    uint32_t getIndex() { return index; }
    uint64_t getBaseAddress() { return baseAddress; }
    uint32_t getSizeInBytes() { return sizeInBytes; }

    void print();

    RawSection* getRawSection();

    void setEntry() { flags |= EntryMask; }
    void setExit()  { flags |= ExitMask; }
    void setTrace() { flags |= TraceMask; }
    void setJumpTable() { flags |= JumpTableMask;}
    void setNoPath() { flags |= NoPathMask; }

    bool isEntry()  { return (flags & EntryMask); }
    bool isExit()   { return (flags & ExitMask); }
    bool isTrace()  { return (flags & TraceMask); }
    bool isJumpTable() { return (flags & JumpTableMask); }
    bool isNoPath() { return (flags & NoPathMask); }

    void setIndex(uint32_t idx);

    void setSizeInBytes(uint32_t size) { sizeInBytes = size; }

    void addEdge(BasicBlock* to);

    XCoffFile* getXCoffFile();

    void findMemoryFloatOps();

    uint32_t getInstructionCount() { return (sizeInBytes/sizeof(PowerPCInstruction)); }

    HashCode getHashCode() { return hashCode; }

    void setSources(uint32_t n,BasicBlock** arr) { numOfSources = n; sources = arr; }
    void setTargets(uint32_t n,BasicBlock** arr) { numOfTargets = n; targets = arr; }
    
    uint32_t getNumOfMemoryOps() { return numOfMemoryOps; }
    uint32_t getNumOfFloatPOps() { return numOfFloatPOps; }
    MemoryOperation* getMemoryOp(uint32_t idx) { 
        if(idx < numOfMemoryOps) 
            return memoryOps[idx]; 
        return NULL; 
    }

    bool findFirstInstPoint(uint64_t* addr);

    uint32_t getNumOfSources() { return numOfSources; }
    uint32_t getNumOfTargets() { return numOfTargets; }
    BasicBlock* getSourceBlock(uint32_t idx) { ASSERT(idx >= 0 && idx < numOfSources); return sources[idx]; }
    BasicBlock* getTargetBlock(uint32_t idx) { ASSERT(idx >= 0 && idx < numOfTargets); return targets[idx]; }

    bool isDominatedBy(BasicBlock* bb);
    BasicBlock* getImmDominator() { return immDominatedBy; }
    void setImmDominator(BasicBlock* bb) { immDominatedBy = bb; }
};

class FlowGraph {
protected:

    Function* function;

    uint32_t numOfBasicBlocks;
    BasicBlock** basicBlocks; /** sorted according to the base address,index is basic block id**/

    uint32_t numberOfLoops;
    Loop** loops;

public:
    FlowGraph(Function* f) : function(f),numOfBasicBlocks(0),basicBlocks(NULL),numberOfLoops(0),loops(NULL) {}
    ~FlowGraph() {}

    RawSection* getRawSection();
    void print();
    uint32_t getIndex();
    Function* getFunction() { return function; }

    void assignSequenceNumbers();
    uint32_t initializeAllBlocks(BasicBlock** blockArr,BasicBlock* traceBlock,uint32_t arrCount);

    XCoffFile* getXCoffFile();

    void findMemoryFloatOps();

    BasicBlock** getAllBlocks() { return basicBlocks; }
    uint32_t getAllBlocks(BasicBlock** arr);

    BitSet<BasicBlock*>* newBitSet();

    BasicBlock* getBlock(uint32_t idx) { return (idx < numOfBasicBlocks ? basicBlocks[idx] : NULL); }

    uint32_t getNumOfBasicBlocks() { return numOfBasicBlocks; }
    uint32_t getNumOfMemoryOps(); 
    uint32_t getNumOfFloatPOps();
    BasicBlock* getEntryBlock();
    Loop* getLoop(uint32_t idx) { ASSERT(idx >= 0 && idx < numberOfLoops); return loops[idx]; }
    uint32_t getNumberOfLoops() { return numberOfLoops; }
    void buildLoops();
    void printInnerLoops();

    void depthFirstSearch(BasicBlock* root,BitSet<BasicBlock*>* visitedSet,bool set,
                          BitSet<BasicBlock*>* completedSet=NULL,LinkedList<BasicBlock*>* backEdges=NULL);
    void setImmDominatorBlocks(BasicBlock* root=NULL);
};

class Function {
protected:
    RawSection* rawSection;
    uint32_t index;

    uint32_t numOfAllSymbols;
    Symbol** allSymbols;

    uint64_t baseAddress;
    uint32_t sizeInBytes;

    uint32_t insnSizeInBytes;

    FlowGraph* flowGraph;

    HashCode hashCode;

public:

    Function(uint32_t idx,uint32_t symCount,Symbol** symbols,RawSection* sect);
    ~Function() { delete[] allSymbols; }

    void updateSize(uint64_t size) { sizeInBytes = size; }
    void updateInstructionSize();

    AddressIterator getAddressIterator();
    AddressIterator getInstructionIterator();

    uint32_t getIndex() { return index; }

    RawSection* getRawSection() { return rawSection; }
    FlowGraph* getFlowGraph();
    void generateCFG();

    uint32_t getSize() { return sizeInBytes; } 
    uint32_t getInstructionSize() { return insnSizeInBytes; } 

    bool instructionInRange(uint64_t addr) { 
        return ((baseAddress <= addr) && (addr < (baseAddress+insnSizeInBytes))); 
    }

    uint64_t getBaseAddress() { return baseAddress; }

    uint32_t getInstructionCount() { return (insnSizeInBytes/sizeof(PowerPCInstruction)); }
    uint32_t getInstructionIndex(uint64_t addr) { return ((addr-baseAddress)/sizeof(PowerPCInstruction)); }

    void print();

    void parseJumpTable(AddressIterator ait,
                        BasicBlock** allBlocks,BasicBlock* currBlock=NULL,
                        LinkedList<uint32_t>* tgtEdgeBags=NULL,
                        LinkedList<uint32_t>* srcEdgeBags=NULL);

    bool getJumpTableInformation(uint64_t* jumpTableAddr,uint32_t* jumpTableEntryCount,
                                 AddressIterator ait);

    XCoffFile* getXCoffFile();

    void findMemoryFloatOps();

    HashCode getHashCode() { return hashCode; }

    BasicBlock* getBlock(uint32_t idx) { return (flowGraph ? flowGraph->getBlock(idx) : NULL); }

    bool isAnySymbolA(char* name);

    uint32_t getNumOfAllSymbols() { return numOfAllSymbols; }
    uint32_t getAllSymbolNames(char* buffer,uint32_t len,bool onlyFirst);

    void buildLoops() { if (flowGraph) flowGraph->buildLoops(); }

};

#endif
