#ifndef _Instruction_h_
#define _Instruction_h_

#include <Base.h>
#include <PowerPCInstruct.h>
#include <BitSet.h>

class Operand {
protected:
    const static uint32_t IntegerType    = 0x1;
    const static uint32_t GPRegisterType = 0x2;

    uint32_t flags;
    union {
        int32_t iconst;
        uint32_t regist;
    } value;

    Operand(uint32_t flgs) : flags(flgs) {}
    Operand() : flags(0) {}

public:
    ~Operand() {}
    bool isInteger()     { return (flags & IntegerType); }
    bool isGPRegister() { return (flags & GPRegisterType); }
    void print() { 
        if(isInteger()){
            PRINT_INFOR("\t\t(INT %d)",value.iconst); 
        } else if(isGPRegister()){
            PRINT_INFOR("\t\t(GPR %2d)",value.regist);
        }
    }
    int32_t getIconst() { return value.iconst; }
    uint32_t getRegister() { return value.regist; }

    static Operand* GPRegisterOperands;
    static Operand IntegerOperand0; 

    static Operand* initGPRegisterOperands();
};

class IntegerOperand : public Operand {
protected:
public:
    IntegerOperand(int32_t val) : Operand(Operand::IntegerType){
        value.iconst = val;
    }
    ~IntegerOperand() {}
};

class GPRegisterOperand : public Operand {
protected:
public:
    GPRegisterOperand(uint32_t val) : Operand(Operand::GPRegisterType){
        value.regist = val;
    }
    ~GPRegisterOperand() {}
};

class Instruction {
protected:

    static BitSet<>* memoryOperationXops;
    PowerPCInstruction ppci;
public:

    Instruction() { ppci.bits = 0; }
    Instruction(uint32_t bits) { ppci.bits = bits; }
    ~Instruction() {}

    bool isZero() { return !(ppci.bits); }
    uint32_t bits() { return ppci.bits; }

    bool isCondBranch();
    bool isJump();
    bool isReturn();
    bool isCondReturn();
    bool isCall();
    bool isIndirectJump();
    bool isIndirectJumpCtr();
    bool isIndirectJumpLnk();
    bool isOtherBranch();
    bool isAddBeforeJump();
    bool isLoadBeforeJump();

    bool definesLeaders(){ 
        return (isCondBranch() || isJump() || isReturn() || isCall() || isIndirectJump());
    }
    bool hasTargetAddress() {
        return (isCondBranch() || isJump());
    }

    uint64_t getTargetAddress(uint64_t insnAddr);

    bool definesJTEntryCount();
    uint32_t getJTEntryCount();

    bool definesJTBaseAddress();
    int32_t getJTBaseOffsetTOC();

    bool definesJTBaseAddrIndir();
    int32_t getJTBaseAddrIndirOffset();

    void print(uint64_t insnAddr,bool is64Bit);
    static void print(char* buffer,uint64_t baseAddress,uint32_t sizeInBytes,bool is64Bit);

    uint32_t getJTBaseAddrTarget();
    uint32_t getJTBaseAddrIndirSrc();

    uint32_t getLoadBeforeJumpSrc1();

    bool isMemoryDForm();
    bool isMemoryDsForm();
    bool isMemoryXForm();
    bool isMemoryOperation() { return (isMemoryDForm() || isMemoryDsForm() || isMemoryXForm()); }
    bool isUnhandledMemoryOp();

    bool isFloatAForm();
    bool isFloatXForm();
    bool isFloatPOperation() { return (isFloatAForm() || isFloatXForm()); }

    static BitSet<>* initMemoryOperationXops();

    uint32_t getDFormSrc1();
    uint32_t getDFormTgt();
    int32_t  getDFormImmediate();
    int32_t  getDsFormImmediate();
    uint32_t getXFormSrc1();
    uint32_t getXFormSrc2();
    bool     isMemoryXFormButNoSrc2();
    
    static bool isInJumpInsnRange(uint64_t from,uint64_t to);
    static bool isInLoadOffsetInsnRange(int32_t value);

    static Instruction generateCompare(uint32_t reg1,uint32_t reg2,uint32_t field);

    static Instruction generateJumpInsn(uint64_t from,uint64_t to);
    static Instruction generateCallToCTR();
    static Instruction generateReturnToLnk();
    static Instruction generateCallToImmediate(uint64_t from,uint64_t to);
    static Instruction generateCondBranch(uint32_t field,uint32_t op,uint32_t tf,int32_t distance);

    static Instruction generateMoveReg(uint32_t from,uint32_t to);

    static Instruction multiplyImmediate(uint32_t tgt,uint32_t src,int32_t imm);

    static Instruction generateAdd(uint32_t tgt,uint32_t src1,uint32_t src2);
    static Instruction generateAddImm(uint32_t tgt,uint32_t src,int32_t value);
    static Instruction generateIncrement(uint32_t reg,int32_t value);
    static Instruction generateLoadImmediate(uint32_t reg,int32_t value);
    static Instruction generateAddImmShifted(uint32_t tgt,uint32_t src,int32_t imm);
    static Instruction generateLoad32BitHigh(uint32_t reg,int32_t value);

    static Instruction generateOrImm(uint32_t tgt,uint32_t src,int32_t imm);
    static Instruction generateLoad32BitLow(uint32_t reg,int32_t value);

    static Instruction generateXorImm(uint32_t tgt,uint32_t src,int32_t imm);
    static Instruction generateXorImmShifted(uint32_t tgt,uint32_t src,int32_t imm);

    static Instruction generateStoreDouble(uint32_t src,uint32_t base,int32_t offset);
    static Instruction generateLoadDouble(uint32_t tgt,uint32_t base,int32_t offset);
    static Instruction generateLoadWord(uint32_t tgt,uint32_t base,int32_t offset);
    static Instruction generateStoreWord(uint32_t src,uint32_t base,int32_t offset);
    static Instruction generateLoadWordIndx(uint32_t tgt,uint32_t base1,uint32_t base2);
    static Instruction generateStoreWordIndx(uint32_t src,uint32_t base1,uint32_t base2);
    static Instruction generateLoadDoubleIndx(uint32_t tgt,uint32_t base1,uint32_t base2);
    static Instruction generateStoreDoubleIndx(uint32_t src,uint32_t base1,uint32_t base2);

    static Instruction generateStoreDoubleFloat(uint32_t src,uint32_t base,int32_t offset);
    static Instruction generateLoadDoubleFloat(uint32_t tgt,uint32_t base,int32_t offset);
    static Instruction generateStoreWordFloat(uint32_t src,uint32_t base,int32_t offset);
    static Instruction generateLoadWordFloat(uint32_t tgt,uint32_t base,int32_t offset);

    static Instruction generateMoveToSPR(uint32_t tgt,uint32_t regcode);
    static Instruction generateMoveFromSPR(uint32_t tgt,uint32_t regcode);
    static Instruction generateMoveFromCR(uint32_t tgt);
    static Instruction generateMoveToCR(uint32_t src);
    static Instruction generateMoveFromFPSCR(uint32_t tgt);
    static Instruction generateMoveToFPSCR(uint32_t src);

    static Instruction generateSPIncrementWord(int32_t offset);
    static Instruction generateSPIncrementDouble(int32_t offset);

    static Instruction generateAnd(uint32_t tgt,uint32_t src1,uint32_t src2);

    bool isMemoryDFormFloat();
};

#endif
