#ifndef _XCoffFile_h_
#define _XCoffFile_h_

#include <Base.h>
#include <BinaryFile.h>

class FileHeader;
class AOutHeader;
class SectHeader;
class RawSection;
class SymbolTable;
class StringTable;
class DebugSection;
class HashCode;
class BasicBlock;
class RelocationTable;
class LineInfoTable;

class XCoffFile {
private:
    bool is64BitFlag;

    char*        xcoffFileName;

    BinaryInputFile   binaryInputFile;

    FileHeader*  fileHeader;
    AOutHeader*  aOutHeader;
    SectHeader** sectHeaders;
    RawSection** rawSections;

    SymbolTable* symbolTable;
    StringTable* stringTable;
    
    uint32_t numberOfSections;

    uint16_t textSectionIndex;
    uint16_t dataSectionIndex;
    uint16_t bssSectionIndex;
    uint16_t loaderSectionIndex;
    uint16_t tocSectionIndex;
    uint16_t entrySectionIndex;

    uint32_t numberOfFunctions;
    uint32_t numberOfBlocks;
    uint32_t numberOfMemoryOps;
    uint32_t numberOfFloatPOps;
    
    void readFileHeader();
    void readAuxilaryHeader();
    void readSectionHeaders();
    void processOverflowSections();
    void readRawSectionData();
    void readSymbolStringTable(DebugSection* dbg);
    void readRelocLineInfoTable();


    void findFunctions();
    void generateCFGs();
    void findMemoryFloatOps();

public:

    XCoffFile(char* f): is64BitFlag(false),xcoffFileName(f),
                  fileHeader(NULL),aOutHeader(NULL),sectHeaders(NULL),rawSections(NULL),
                  symbolTable(NULL),stringTable(NULL),numberOfSections(0),
                  textSectionIndex(0),dataSectionIndex(0),bssSectionIndex(0),
                  loaderSectionIndex(0),tocSectionIndex(0),entrySectionIndex(0),
                  numberOfFunctions(0),numberOfBlocks(0),numberOfMemoryOps(0),numberOfFloatPOps(0) {}

    ~XCoffFile() { }

    bool is64Bit() { return is64BitFlag; }

    void parse();

    void briefPrint();
    void print();
    void displaySymbols();

    RawSection* findRawSection(uint64_t addr);
    uint64_t    getTOCAddress();
    RawSection* getTOCSection();
    uint32_t getTOCSectionIndex() { return tocSectionIndex; }
    uint64_t    readTOC(int32_t offset);
    char* getXCoffFileName() { return xcoffFileName; }

    uint32_t      getNumberOfSections() { return numberOfSections; }

    FileHeader*  getFileHeader() { return fileHeader; }
    AOutHeader*  getAOutHeader() { return aOutHeader; }
    SymbolTable* getSymbolTable() { return symbolTable; }
    StringTable* getStringTable() { return stringTable; }

    SectHeader* getSectHeader(uint32_t idx) { return sectHeaders[idx]; }
    RawSection* getRawSection(uint32_t idx) { return rawSections[idx]; }
    RelocationTable* getRelocationTable(uint32_t idx);
    LineInfoTable* getLineInfoTable(uint32_t idx);

    uint32_t getNumberOfFunctions() { return numberOfFunctions; }
    uint32_t getNumberOfBlocks()    { return numberOfBlocks; }
    uint32_t getNumberOfMemoryOps() { return numberOfMemoryOps; }
    uint32_t getNumberOfFloatPOps() { return numberOfFloatPOps; }

    BasicBlock* findBasicBlock(HashCode* hashCode);
    uint32_t getAllBlocks(BasicBlock** arr);

    uint32_t getFileSize();

    uint16_t getLoaderSectionIndex() { return loaderSectionIndex; }
    RawSection* getLoaderSection();
    uint16_t getDataSectionIndex() { return dataSectionIndex; }
    uint64_t getDataSectionVAddr();
    uint32_t getDataSectionSize();
    RawSection* getBSSSection();
    uint16_t getBSSSectionIndex() { return bssSectionIndex; }
    uint64_t getBSSSectionVAddr();
    uint64_t getTextSectionIndex() { return textSectionIndex; }
    uint64_t getTextSectionVAddr();

    void setLineInfoFinder();
    void findLoops();

    void testBitSet();

};

/**
    _________________________________________   INCLUDE FILE
    |_______________HEADER_DATA_____________|
    |                                       |
    |   File Header                         |   "filehdr.h"
    |.......................................|
    |                                       |
    |   Auxilliary Header Information       |   "aouthdr.h"
    |_______________________________________|
    |   ".text" section header              |   "scnhdr.h"
    |.......................................|
    |   ".data" section header              |   ''
    |.......................................|
    |   ".bss" section header               |   ''
    |.......................................|
    |   ".loader" section header            |   ''
    |.......................................|
    |   ".typchk" section header            |   ''
    |.......................................|
    |   ".debug" section header             |   ''
    |.......................................|
    |   ".except" section header            |   ''
    |.......................................|
    |   ".info" section header              |   ''
    |_______________________________________|
    |______________RAW_DATA_________________|
    |   ".text" section data                |
    |.......................................|
    |   optional ".pad" section data        |
    |       (file alignment)                |
    |.......................................|
    |   ".data" section data                |
    |.......................................|
    |   ".loader" section data              |   "loader.h"
    |.......................................|
    |   ".typchk" section data              |   "typchk.h"
    |.......................................|
    |   ".debug" section data               |   "dbug.h"
    |.......................................|
    |   ".except" section data              |   "exceptab.h"
    |.......................................|
    |   ".info" section data                |
    |_______________________________________|
    |____________RELOCATION_DATA____________|
    |                                       |
    |   ".text" section relocation data     |   "reloc.h"
    |                                       |
    |.......................................|
    |                                       |
    |   ".data" section relocation data     |   ''
    |                                       |
    |_______________________________________|
    |__________LINE_NUMBER_DATA_____________|
    |                                       |
    |   ".text" section line numbers        |   "linenum.h"
    |                                       |
    |_______________________________________|
    |________________SYMBOL_TABLE___________|
    |                                       |
    |   ".text", ".data" and ".bss"         |   "syms.h"
    |       section symbols                 |   "storclass.h"
    |                                       |
    |_______________________________________|
    |________________STRING_TABLE___________|
    |                                       |
    |   long symbol names                   |
    |_______________________________________|

**/

#endif /* _XCoffFile_h_ */
