#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <AddressSpaceCoverage.h>

/* use a two level hierahical table to keep track of pages */
const uint32_t page_size_bits = 12;
const uint32_t level1_bits = 14;
const uint32_t level2_bits = 6;

const uint32_t level1_shift = 18;
const uint32_t level2_shift = 12;

static uint32_t level1_mask;
static uint32_t level2_mask;
static uint32_t num_level1_entries;
static uint32_t num_level2_entries;
static uint32_t num_level2_entries2;

Counter_t **level1_Table = 0;
Counter_t totalVisits = 0;
Counter_t totalEventSamples = 0;
Counter_t totalAddressSamples = 0;

uint32_t distinctPagesVisited = 0;
uint32_t allocatedLevel2Tables = 0;

/** Sample by memory address. 
    Consider memory organized in blocks (think cache lines). 
    Sampling works at block granularity. We consider all accesses to a 
    selected block. We ignore all accesses to not-selected blocks. 
    Least significant block_size_bits define an offset into a block.
    Next block_sampling_bits define the mask we use to select a block. 
    Thus, only one out of 2^block_sampling_bits is selected. In one page we
    have 2^(page_size_bits - block_size_bits - block_sampling_bits) blocks 
    that can be selected. 
    page_size_bits >= block_sampling_bits+block_size_bits must hold true.
 */
const uint32_t block_size_bits = 6;
const uint32_t block_sampling_bits = 4;  /* 1 in 16 blocks is selected*/
static uint32_t block_index_shift;
static uint32_t block_index_mask;

/*
static uint32_t event_sampling_interval = 0;
*/

FILE* openOutputFile(const char* comment){

    uint32_t rankId = 0;

    char strBuffer[__MAX_STRING_SIZE];

    uint32_t processId = getpid();
    uint32_t hostId = gethostid();

    FILE* fp = fopen("RankPid","r");
    if(fp){
        uint32_t entryFound = 0;
        while(fgets(strBuffer,1024,fp)){
            strBuffer[1023] = '\0';
            uint32_t corresPid = 0,corresHid = 0;
            sscanf(strBuffer,"%u %u %u",&rankId,&corresPid,&corresHid);
            if((corresPid == processId) && (corresHid == hostId)){
                entryFound = 1;
                break;
            }
        }
        fclose(fp);
        if(!entryFound){
            rankId = processId;
        }
    } else {
        rankId = processId;
    }

    char appName[__MAX_STRING_SIZE];
    uint32_t phaseId = 0;
    char extension[__MAX_STRING_SIZE];

    sscanf(comment,"%s %u %s",appName,&phaseId,extension);

    if(phaseId > 0){
        sprintf(strBuffer,"%s.phase.%d.meta_%04d.%s",appName,phaseId,rankId,extension);
    } else {
        sprintf(strBuffer,"%s.meta_%04d.%s",appName,rankId,extension);
    }
    
    fp = fopen(strBuffer,"w");
    if(!fp){
        fprintf(stderr,"Error : Can not open %s to write results\n",strBuffer);
        return NULL;
    }

    fprintf(fp,"# appname   = %s\n",appName);
    fprintf(fp,"# extension = %s\n",extension);
    fprintf(fp,"# phase     = %d\n",phaseId);
    fprintf(fp,"# rank      = %d\n",rankId);

    return fp;
}

/*** This method is called by the instrumentation when the address buffer
     is full. The address buffer is a list of records of type AddressEntry.
     Each AddressEntry contains address, basic block sequence id, memory instruction
     index within the block of each memory access. The first entry in the address
     stream buffer entry does not contain a valid address. Instead, the blockId
     (also defined as lastFreeIdx) contains valid data and it points to the first index
     in the buffer which is empty. Thus, at the end of simulation this field has
     to be set to 1 (addressStream->lastFreeIdx = 1) to indicate buffer is processed.
     Moreover, the valid data is in the indecies from 1 to (addressStream->lastFreeIdx-1).

     Note that for 32-bit and 64-bit executables, the size of virtiual address changes
     thus the address type is set to be 32-bit or 64-bit unsigned integer depending on
     the executable. The macro METASIM_32_BIT_LIB can be used as in NewCacheSimulator.h
     for correct representation of 32bit and 64bit libraries.

     arguments
     @ char* base : pointer to the beginning of address stream buffer. each entry is
                    a record of AddressEntry. 
     @ uint32_t entryCount : size of the address stream buffer. Note that it is not
                    the number of valid entries. It is the size of buffer in elements.
     @ const char* comment : a string contains additional information passed from the
                    executable to this function (name of the exec, suffix, and so on).
***/

void MetaSim_simulFuncCall_Simu(char* base,uint32_t entryCount,const char* comment){

    AddressEntry* addressStream = (AddressEntry*)base;
    Attribute_t   startIndex = 1;
    Attribute_t   lastIndex = addressStream->lastFreeIdx;
    lastIndex--;
/*
    printf("Address stream will be processed here. Args are [0x%x] [%d] [%s] -- last entry at (%d)\n",
            base,entryCount,comment,lastIndex); 
 */
    processSamples(addressStream,startIndex,lastIndex);

    /*** End of simulation. Set the lastFreeIdx to 1 to 
         set address stream buffer to be empty. ***/
    addressStream->lastFreeIdx = 1;
}

/*** This method is called at program termination. So the results of
     simulations is printed in this method. Arguments are the same as in 
     MetaSim_simulFuncCall_Simu function above.
     The users are expected to include their code in this module
     to save/print/post-process simulation results, as shown below.
***/

void MetaSim_endFuncCall_Simu(char* base,uint32_t entryCount,const char* comment){

    /*** Simulate the remaining address stream
         (since last call to MetaSim_simulFuncCall_Simu) ***/
    MetaSim_simulFuncCall_Simu(base,entryCount,comment);

    /*** Process/print your cache simulation results (termination)
              |  |   |   | 
              V  V   V   V  
    ****/
    uint32_t i = 0;
    FILE* fd = openOutputFile(comment);
    if (fd)
    {
       float eventSamplingFactor = ((float)totalVisits)/totalEventSamples;
       uint32_t addressSamplingFactor = 1 << block_sampling_bits;
       fprintf (fd, "# totalVisits           = %lld\n", totalVisits);
       fprintf (fd, "# eventSamples          = %lld\n", totalEventSamples);
       fprintf (fd, "# addressSamples        = %lld\n", totalAddressSamples);
       fprintf (fd, "# distinctPages         = %d\n", distinctPagesVisited);
       fprintf (fd, "# level 2 tables        = %d\n", allocatedLevel2Tables);
       fprintf (fd, "# eventSamplingFactor   = %g\n", eventSamplingFactor);
       fprintf (fd, "# addressSamplingFactor = %d\n", addressSamplingFactor);
       fprintf (fd, "# pageIdx\tnoSampling\teventSampling\taddressSampling\n");
       if (level1_Table)
          for (i=0 ; i<num_level1_entries ; ++i)
             if (level1_Table[i])
             {
                Address_t basePageIdx = i << level2_shift;
                Counter_t *level2_Table = level1_Table[i];
                uint32_t j;
                for (j=0 ; j<num_level2_entries ; ++j)
                   if (level2_Table[j])   /* found page with count > 0 */
                      fprintf (fd, "%d\t%lld\t%lld\t%lld\n", basePageIdx+j, 
                          level2_Table[j], 
                          (Counter_t)(level2_Table[j+num_level2_entries]*eventSamplingFactor),
                          level2_Table[j+num_level2_entries2]*addressSamplingFactor);
             }
    }
/*    printf("Application termination. Print your simulation results\n");
 */
}

/*** This method needs to be implemented by users according to their simulation needs.
     Currently the method goes over each address in the stream and stores block and memory
     operation ids.
***/

void processSamples(AddressEntry* addressStream,Attribute_t startIndex,Attribute_t lastIndex)
{
    /*** Your cache simulation methodology (For the given address sequence)
                   |  |   |   |
                   V  V   V   V
    ***/
    
    uint32_t i = 0;
    Attribute_t samplingStartIdx = startIndex;
    Attribute_t samplingLastIdx = lastIndex;
    if (!level1_Table)
    {
       level1_mask = (1<<level1_bits)-1;
       level2_mask = (1<<level2_bits)-1;
       num_level1_entries = 1<<level1_bits;
       num_level2_entries = 1<<level2_bits;
       block_index_shift = block_size_bits;
       block_index_mask = (1<<block_sampling_bits)-1;
       level1_Table = (Counter_t**) malloc (num_level1_entries*sizeof (Counter_t*));
       bzero (level1_Table, num_level1_entries*sizeof (Counter_t*));
       num_level2_entries2 = num_level2_entries + num_level2_entries;
    }

    /* startIndex is always 1 here */
    totalVisits += lastIndex;   /* (lastIndex - startIndex + 1); */
    
    if(currentSamplingStatus == ignoring_accesses){
        alreadyIgnored += lastIndex;
        if(alreadyIgnored > __WHICH_IGNORING_VALUE){
            currentSamplingStatus = sampling_accesses;
            alreadySampled = alreadyIgnored - __WHICH_IGNORING_VALUE;
#ifdef EXTENDED_SAMPLING
            if(alreadySampled > lastIndex){
                samplingStartIdx = 1;
            } else {
#endif
            samplingStartIdx = (lastIndex - alreadySampled + 1);
#ifdef EXTENDED_SAMPLING
            }
            Attribute_t tmp_rand = rand_value;
            rand_value = (Attribute_t)(random() % (SEGMENT_COUNT+1));
#ifdef DEBUG_RUN
            fprintf(stdout,"* %12lld IGNORED > %8lld sampled %8lld ignored %8lld [%6d,%6d] rand (%2d,%2d)\n",
                    totalVisits,__WHICH_IGNORING_VALUE,alreadySampled,alreadyIgnored,
                    samplingStartIdx,samplingLastIdx,tmp_rand,rand_value);
#endif
            __WHICH_IGNORING_VALUE = __IGNORING_INTERVAL_MAX - (tmp_rand*(__IGNORING_INTERVAL_MAX/SEGMENT_COUNT));
            __WHICH_IGNORING_VALUE += (Counter_t)(rand_value*(__IGNORING_INTERVAL_MAX/SEGMENT_COUNT));
#endif
            alreadyIgnored = 0;
        }
    } else {
        alreadySampled += lastIndex;
        if(alreadySampled > __WHICH_SAMPLING_VALUE){
            currentSamplingStatus = ignoring_accesses;
            alreadyIgnored = alreadySampled - __WHICH_SAMPLING_VALUE;
#ifdef EXTENDED_SAMPLING
            if(alreadyIgnored > lastIndex){
                samplingLastIdx = 0;
            } else {
#endif
            samplingLastIdx = (lastIndex - alreadyIgnored);
#ifdef EXTENDED_SAMPLING
            }
            Attribute_t tmp_rand = (random() % (SEGMENT_COUNT+1)) + (SEGMENT_COUNT/2);
#ifdef DEBUG_RUN
            fprintf(stdout,"- %12lld SAMPLED > %8lld ignored %8lld sampled %8lld [%6d,%6d] rand (%2d)\n",
                    totalVisits,__WHICH_SAMPLING_VALUE,alreadyIgnored,alreadySampled,
                    samplingStartIdx,samplingLastIdx,tmp_rand);
#endif
            __WHICH_SAMPLING_VALUE = (Counter_t)((1.0*tmp_rand/SEGMENT_COUNT)*__SAMPLING_INTERVAL_MAX);
#endif
            alreadySampled = 0;
        }
    }
    if (currentSamplingStatus == sampling_accesses)
       totalEventSamples += (samplingLastIdx + 1 - samplingStartIdx);
    
    /*** Iterate over the address entries in the address stream ***/
    for (i=startIndex ; i<=lastIndex ; ++i){
        /*** Each entry contains virtual address (32 bit or 64 bit depending
             on the exceutable's being 32bit or 64bit), sequence id of the block
             which requested the access and the memory operation index within the block
             that requested the address ***/

        AddressEntry* currentEntry = (addressStream + i);
        Attribute_t   blockSeqId = currentEntry->blockId;
        Attribute_t   memOpIdx = currentEntry->memOpId;
        Address_t     location = currentEntry->address;
        
        uint32_t level1_entry = (location >> level1_shift) & level1_mask;
        uint32_t level2_entry = (location >> level2_shift) & level2_mask;
        Counter_t **level2_Table = level1_Table + level1_entry;
        if (! (*level2_Table))
        {
           *level2_Table = (Counter_t*) malloc (num_level2_entries*Total_SamplingMethods*
                       sizeof (Counter_t));
           bzero (*level2_Table, num_level2_entries*Total_SamplingMethods*sizeof (Counter_t));
           allocatedLevel2Tables += 1;
        }
        if (! (*level2_Table)[level2_entry]) 
           distinctPagesVisited += 1;
        (*level2_Table)[level2_entry] += 1;
        
        /** check if we consider this block under the event sampling rule */
        if (currentSamplingStatus==sampling_accesses && 
                samplingStartIdx<=i && i<=samplingLastIdx)
        {
           (*level2_Table)[level2_entry+num_level2_entries] += 1;
        }

        /** check if we consider this block under the address sampling rule */
        if (((location >> block_index_shift) & block_index_mask) == 0)
        {
           (*level2_Table)[level2_entry+num_level2_entries2] += 1;
           totalAddressSamples += 1;
        }

        /*** Now we have an address and its requestor basic block and memory op
             id. Simulate using your own modules.***/
    }
}
