#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>

#include <CompareCacheSimulations.h>

#define STATUS_IDX(__sys,__smpl,__lvl,__sta) ((__sys)*(Total_SamplingType,__MAX_LEVEL*Total_AccessStatus) + (__smpl)*(__MAX_LEVEL*Total_AccessStatus) + (__lvl) * Total_AccessStatus + (__sta))

#ifdef USE_SATURATION
Counter_t       lastSaturationAccess = 0;
Counter_t       saturatedBlockCount = 0;
#endif
Counter_t       maxProcessedVisits = 0;
Counter_t       totalNumberOfAccesses[Total_SamplingType];
/*
Counter_t       addressAccesses[__SYSTEM_COUNT*__MAX_LEVEL];
*/

const uint32_t set_sampling_bits = 3;  /* 1 in 8 blocks is selected*/
uint32_t set_sampling_mask;

uint32_t        blockCount = 0;
BasicBlockInfo* blocks = NULL;

const char*
samplingName (SamplingType stype)
{
   switch (stype)
   {
      case no_sampling:      return ("NoSampling"); break;
      case event_sampling:   return ("EventSampling"); break;
      case address_sampling: return ("AddressSampling"); break;
      default: return ("UnknownSampling"); break;
   }
   return (0);
}

static unsigned int SEED = 93186752;
int getRandomNumber()
{
    double x;
    static unsigned a = 1588635695, m = 4294967291U, q = 2, r = 1117695901;
    SEED = a*(SEED % q) - r*(SEED / q);
    x = (double)SEED / (double)m; /* return the next random number x: 0 <= x < 1 */

    return ( (int)128*x +1);
}

FILE* openOutputFile(const char* comment){

    uint32_t rankId = 0;

    char strBuffer[__MAX_STRING_SIZE];

    uint32_t processId = getpid();
    uint32_t hostId = gethostid();

    FILE* fp = fopen("RankPid","r");
    if(fp){
        uint32_t entryFound = 0;
        while(fgets(strBuffer,1024,fp)){
            strBuffer[1023] = '\0';
            uint32_t corresPid = 0,corresHid = 0;
            sscanf(strBuffer,"%u %u %u",&rankId,&corresPid,&corresHid);
            if((corresPid == processId) && (corresHid == hostId)){
                entryFound = 1;
                break;
            }
        }
        fclose(fp);
        if(!entryFound){
            rankId = processId;
        }
    } else {
        rankId = processId;
    }

    char appName[__MAX_STRING_SIZE];
    uint32_t phaseId = 0;
    char extension[__MAX_STRING_SIZE];

    sscanf(comment,"%s %u %s",appName,&phaseId,extension);

    if(phaseId > 0){
        sprintf(strBuffer,"%s.phase.%d.meta_%04d.%s",appName,phaseId,rankId,extension);
    } else {
        sprintf(strBuffer,"%s.meta_%04d.%s",appName,rankId,extension);
    }
    
    fp = fopen(strBuffer,"w");
    if(!fp){
        fprintf(stderr,"Error : Can not open %s to write results\n",strBuffer);
        return NULL;
    }

    fprintf(fp,"# appname   = %s\n",appName);
    fprintf(fp,"# extension = %s\n",extension);
    fprintf(fp,"# phase     = %d\n",phaseId);
    fprintf(fp,"# rank      = %d\n",rankId);

    return fp;
}

void 
initCaches() 
{
    register uint32_t systemIdx;
    register uint32_t level;
    register uint32_t setIdx;
    
    for (level=0 ; level<Total_SamplingType ; ++level)
       totalNumberOfAccesses[level] = 0;
    set_sampling_mask = (1 << set_sampling_bits) - 1;
    
    for (systemIdx=0 ; systemIdx<systemCount ; systemIdx++) {
        register MemoryHierarchy* memoryHierarchy = (systems + systemIdx);
        register uint32_t totalLineCount = 0;
        for (level=0 ; level<memoryHierarchy->levelCount ; level++) 
        {
            register Attribute_t setCount = memoryHierarchy->levels[level].attributes[number_of_sets];
            register Attribute_t linePerSet = memoryHierarchy->levels[level].attributes[set_associativity];
            register Attribute_t sizeInLines = setCount*linePerSet;
            totalLineCount += sizeInLines;
        }

        Address_t* allLines = (Address_t*)malloc (sizeof(Address_t)*totalLineCount*Total_SamplingType);
        bzero(allLines,sizeof(Address_t)*totalLineCount*Total_SamplingType);

        for(level=0;level<memoryHierarchy->levelCount;level++){

            register Attribute_t setCount = memoryHierarchy->levels[level].attributes[number_of_sets];
            register Attribute_t linePerSet = memoryHierarchy->levels[level].attributes[set_associativity];
            register Attribute_t sizeInLines = setCount*linePerSet;

            for (SamplingType stype=0 ; stype<Total_SamplingType ; ++stype) {
               memoryHierarchy->levels[level].hitMissCounters[stype*Total_AccessStatus+cache_hit] = 0;
               memoryHierarchy->levels[level].hitMissCounters[stype*Total_AccessStatus+cache_miss] = 0;
            }
            memoryHierarchy->levels[level].content = allLines;
#ifdef PER_SET_RECENT
            memoryHierarchy->levels[level].mostRecent = (uint16_t*)malloc (sizeof(uint16_t)*setCount*Total_SamplingType);
            for (setIdx=0 ; setIdx<setCount*Total_SamplingType ; ++setIdx) {
                memoryHierarchy->levels[level].mostRecent[setIdx] = (linePerSet-1);
            }
#else
            for (SamplingType stype=0 ; stype<Total_SamplingType ; ++stype)
                memoryHierarchy->levels[level].mostRecent[stype] = (linePerSet-1);
#endif

            allLines += sizeInLines*Total_SamplingType;
        }
    }
}

#ifndef bool
typedef enum {
   false,
   true
} bool;
#endif

void 
processSamples (BufferEntry* entries,Attribute_t startIndex,Attribute_t lastIndex) 
{
    register uint32_t systemIdx;
    register uint32_t accessIdx;
    register uint32_t level;
    bool firstPass = true;

    for (systemIdx=0 ; systemIdx<systemCount ; ++systemIdx) {
        register MemoryHierarchy* memoryHierarchy = (systems + systemIdx);

        for (accessIdx=startIndex ; accessIdx<=lastIndex ; ++accessIdx) {
            register BufferEntry* currentEntry = (entries + accessIdx);
            register Attribute_t currentMemOp = currentEntry->memOpId;
            register Address_t currentAddress = currentEntry->address;

            register BasicBlockInfo* currentBlock = (blocks + currentEntry->blockId);
            if (firstPass)
               currentBlock->visitCount[no_sampling] += 1;

            
            register uint8_t levelCount = memoryHierarchy->levelCount;
            for (level=0 ; level<levelCount ; ++level) {
                register Cache* cache = &(memoryHierarchy->levels[level]);
                register Attribute_t sizeInBits = cache->attributes[line_size_in_bits];
                register Attribute_t numOfSets  = cache->attributes[number_of_sets];
                register Attribute_t assocCount = cache->attributes[set_associativity];

                register uint32_t clineIdx = currentAddress >> sizeInBits;
                register uint32_t setIdx = (clineIdx % numOfSets);
                
                /* process this memory access using no_sampling */
                register Address_t* content = cache->content + (numOfSets*no_sampling) + (setIdx * assocCount);

                register uint8_t mostRecent = 0;
                register uint32_t i = 0;
                register AccessStatus status = cache_miss;
                for(i=0;i<assocCount;i++){
                    if(content[i] == clineIdx){
                        status = cache_hit;
                        mostRecent = i;
                        break;
                    }
                }
                if(status == cache_hit){
                    cache->hitMissCounters[no_sampling*Total_AccessStatus+cache_hit]++;
                    currentBlock->hitMissCounters[STATUS_IDX(systemIdx,no_sampling,level,cache_hit)]++;
                } else {
                    cache->hitMissCounters[no_sampling*Total_AccessStatus+cache_miss]++;
                    currentBlock->hitMissCounters[STATUS_IDX(systemIdx,no_sampling,level,cache_miss)]++;

                    register Attribute_t replPolicy = cache->attributes[replacement_policy];
                    if (replPolicy == repl_ran) {
                        mostRecent = getRandomNumber();
                    } else if (replPolicy == repl_lru) {
#ifdef PER_SET_RECENT
                        mostRecent = (cache->mostRecent[setIdx*Total_SamplingType+no_sampling] + 1) % assocCount;
#else
                        mostRecent = (cache->mostRecent[no_sampling] + 1) % assocCount;
#endif
                    } else if (replPolicy == repl_dir) {
                        mostRecent = 0;
                    }

                    content[mostRecent] = clineIdx;
                }
#ifdef PER_SET_RECENT
                cache->mostRecent[setIdx*Total_SamplingType+no_sampling] = mostRecent;
#else
                cache->mostRecent[no_sampling] = mostRecent;
#endif

                /* process this memory access using address_sampling 
                 * we can think of address sampling as setIdx sampling; We simulate
                 * only a fraction of the sets.
                 */
                if (!(setIdx & set_sampling_mask))
                {
                   /*
                   currentBlock->addressCount[systemIdx*__MAX_LEVEL+level] += 1;
                   addressAccesses[systemIdx*__MAX_LEVEL+level] += 1;
                   */
                   
                   register Address_t* content = cache->content + (numOfSets*address_sampling) + (setIdx * assocCount);

                   register uint8_t mostRecent = 0;
                   register uint32_t i = 0;
                   register AccessStatus status = cache_miss;
                   for(i=0;i<assocCount;i++){
                       if(content[i] == clineIdx){
                           status = cache_hit;
                           mostRecent = i;
                           break;
                       }
                   }
                   if(status == cache_hit){
                       cache->hitMissCounters[address_sampling*Total_AccessStatus+cache_hit]++;
                       currentBlock->hitMissCounters[STATUS_IDX(systemIdx,address_sampling,level,cache_hit)]++;
                   } else {
                       cache->hitMissCounters[address_sampling*Total_AccessStatus+cache_miss]++;
                       currentBlock->hitMissCounters[STATUS_IDX(systemIdx,address_sampling,level,cache_miss)]++;

                       register Attribute_t replPolicy = cache->attributes[replacement_policy];
                       if (replPolicy == repl_ran) {
                           mostRecent = getRandomNumber();
                       } else if (replPolicy == repl_lru) {
#ifdef PER_SET_RECENT
                           mostRecent = (cache->mostRecent[setIdx*Total_SamplingType+address_sampling] + 1) % assocCount;
#else
                           mostRecent = (cache->mostRecent[address_sampling] + 1) % assocCount;
#endif
                       } else if (replPolicy == repl_dir) {
                           mostRecent = 0;
                       }

                       content[mostRecent] = clineIdx;
                   }
#ifdef PER_SET_RECENT
                   cache->mostRecent[setIdx*Total_SamplingType+address_sampling] = mostRecent;
#else
                   cache->mostRecent[address_sampling] = mostRecent;
#endif
                }
                if(status == cache_hit){
                    break;
                }
            }
        }
        if (firstPass) 
           firstPass = false;
    }
}

void 
processSamples_EventSampling (BufferEntry* entries, Attribute_t startIndex, Attribute_t lastIndex) 
{
    register uint32_t systemIdx;
    register uint32_t accessIdx;
    register uint32_t level;
    bool firstPass = true;

    for (systemIdx=0 ; systemIdx<systemCount ; systemIdx++) {
        register MemoryHierarchy* memoryHierarchy = (systems + systemIdx);

        for (accessIdx=startIndex ; accessIdx<=lastIndex ; ++accessIdx) {
            register BufferEntry* currentEntry = (entries + accessIdx);
            register Attribute_t currentMemOp = currentEntry->memOpId;
            register Address_t currentAddress = currentEntry->address;

            register BasicBlockInfo* currentBlock = (blocks + currentEntry->blockId);
            if (firstPass)
               currentBlock->visitCount[event_sampling] += 1;

            register uint8_t levelCount = memoryHierarchy->levelCount;
            for (level=0 ; level<levelCount ; ++level) {
                register Cache* cache = &(memoryHierarchy->levels[level]);
                register Attribute_t sizeInBits = cache->attributes[line_size_in_bits];
                register Attribute_t numOfSets  = cache->attributes[number_of_sets];
                register Attribute_t assocCount = cache->attributes[set_associativity];

                register uint32_t clineIdx = currentAddress >> sizeInBits;
                register uint32_t setIdx = (clineIdx % numOfSets);
                
                /* process this memory access using event_sampling */
                register Address_t* content = cache->content + (numOfSets*event_sampling) + (setIdx * assocCount);

                register uint8_t mostRecent = 0;
                register uint32_t i = 0;
                register AccessStatus status = cache_miss;
                for (i=0 ; i<assocCount ; i++) {
                    if (content[i] == clineIdx) {
                        status = cache_hit;
                        mostRecent = i;
                        break;
                    }
                }
                if(status == cache_hit){
                    cache->hitMissCounters[event_sampling*Total_AccessStatus+cache_hit]++;
                    currentBlock->hitMissCounters[STATUS_IDX(systemIdx,event_sampling,level,cache_hit)]++;
                } else {
                    cache->hitMissCounters[event_sampling*Total_AccessStatus+cache_miss]++;
                    currentBlock->hitMissCounters[STATUS_IDX(systemIdx,event_sampling,level,cache_miss)]++;

                    register Attribute_t replPolicy = cache->attributes[replacement_policy];
                    if (replPolicy == repl_ran) {
                        mostRecent = getRandomNumber();
                    } else if (replPolicy == repl_lru) {
#ifdef PER_SET_RECENT
                        mostRecent = (cache->mostRecent[setIdx*Total_SamplingType+event_sampling] + 1) % assocCount;
#else
                        mostRecent = (cache->mostRecent[event_sampling] + 1) % assocCount;
#endif
                    } else if (replPolicy == repl_dir) {
                        mostRecent = 0;
                    }

                    content[mostRecent] = clineIdx;
                }
#ifdef PER_SET_RECENT
                cache->mostRecent[setIdx*Total_SamplingType+event_sampling] = mostRecent;
#else
                cache->mostRecent[event_sampling] = mostRecent;
#endif

                if(status == cache_hit){
                    break;
                }
            }
        }
        if (firstPass)
           firstPass = false;
    }
}

void 
MetaSim_simulFuncCall_Simu (char* base,uint32_t entryCount,const char* comment)
{
    register uint32_t      i;

    register BufferEntry* entries = (BufferEntry*)base;
    register Attribute_t startIndex = 1;
    register Attribute_t lastIndex = entries->lastFreeIdx;
    lastIndex--;

    if (!blocks) {
        char      appName[__MAX_STRING_SIZE];
        uint32_t  phaseId = 0;
        char      extension[__MAX_STRING_SIZE];

        sscanf (comment,"%s %u %s %u",appName,&phaseId,extension,&blockCount);
        blocks = (BasicBlockInfo*)malloc (sizeof (BasicBlockInfo) * blockCount);
        bzero (blocks,sizeof(BasicBlockInfo)*blockCount);
        initCaches ();
        maxProcessedVisits = blockCount*10000000;
    }

    if (totalNumberOfAccesses[no_sampling] > maxProcessedVisits)
    {
        entries->lastFreeIdx = 1;
        return;
    }
    
    totalNumberOfAccesses[no_sampling] += lastIndex;
    processSamples (entries, startIndex, lastIndex);

#ifdef USE_SATURATION
    if (blockCount && (saturatedBlockCount == blockCount)) {
        entries->lastFreeIdx = 1;
        return;
    }

    lastSaturationAccess = totalNumberOfAccesses;
#endif

#ifndef NO_SAMPLING_MODE
    if (currentSamplingStatus == ignoring_accesses) {
        alreadyIgnored += lastIndex;
        if (alreadyIgnored > __WHICH_IGNORING_VALUE) {
            currentSamplingStatus = sampling_accesses;
            alreadySampled = alreadyIgnored - __WHICH_IGNORING_VALUE;
#ifdef EXTENDED_SAMPLING
            if (alreadySampled > lastIndex) {
                startIndex = 1;
            } else {
#endif
            startIndex = (lastIndex - alreadySampled + 1);
#ifdef EXTENDED_SAMPLING
            }
            Attribute_t tmp_rand = rand_value;
            rand_value = (Attribute_t)(random() % (SEGMENT_COUNT+1));
#ifdef DEBUG_RUN
            fprintf (stdout,"* %12lld IGNORED > %8lld sampled %8lld ignored %8lld [%6d,%6d] rand (%2d,%2d)\n",
                    totalNumberOfAccesses[no_sampling],__WHICH_IGNORING_VALUE,alreadySampled,alreadyIgnored,
                    startIndex,lastIndex,tmp_rand,rand_value);
#endif
            __WHICH_IGNORING_VALUE = __IGNORING_INTERVAL_MAX - (tmp_rand*(__IGNORING_INTERVAL_MAX/SEGMENT_COUNT));
            __WHICH_IGNORING_VALUE += (Counter_t)(rand_value*(__IGNORING_INTERVAL_MAX/SEGMENT_COUNT));
#endif
            alreadyIgnored = 0;
        } else {
            entries->lastFreeIdx = 1;
            return;
        }
    } else {
        alreadySampled += lastIndex;
        if(alreadySampled > __WHICH_SAMPLING_VALUE){
            currentSamplingStatus = ignoring_accesses;
            alreadyIgnored = alreadySampled - __WHICH_SAMPLING_VALUE;
#ifdef EXTENDED_SAMPLING
            if(alreadyIgnored > lastIndex){
                lastIndex = 0;
            } else {
#endif
            lastIndex = (lastIndex - alreadyIgnored);
#ifdef EXTENDED_SAMPLING
            }
            Attribute_t tmp_rand = (random() % (SEGMENT_COUNT+1)) + (SEGMENT_COUNT/2);
#ifdef DEBUG_RUN
            fprintf(stdout,"- %12lld SAMPLED > %8lld ignored %8lld sampled %8lld [%6d,%6d] rand (%2d)\n",
                    totalNumberOfAccesses,__WHICH_SAMPLING_VALUE,alreadyIgnored,alreadySampled,
                    startIndex,lastIndex,tmp_rand);
#endif
            __WHICH_SAMPLING_VALUE = (Counter_t)((1.0*tmp_rand/SEGMENT_COUNT)*__SAMPLING_INTERVAL_MAX);
#endif
            alreadySampled = 0;
        }
    }
#endif /* NO_SAMPLING_MODE */

#ifdef EXTENDED_SAMPLING
    if(lastIndex < startIndex){
        entries->lastFreeIdx = 1;
#ifdef DEBUG_RUN
        fprintf(stdout,"DONE lastIndex < startIndex\n");
#endif
        return;
    }
#endif


    totalNumberOfAccesses[event_sampling] += (lastIndex-startIndex+1);

#ifdef USE_SATURATION
    register int32_t lastInvalidEntry = startIndex-1;

    register BasicBlockInfo* previousBlock = NULL;
    for (i=startIndex ; i<=lastIndex ; ++i) {
        register BufferEntry* currentEntry = (entries + i);
        register BasicBlockInfo* currentBlock = (blocks + currentEntry->blockId);
        register Attribute_t currentMemOp = currentEntry->memOpId;

        if (!currentMemOp || (currentBlock != previousBlock)) {
            currentBlock->visitCount++;
            previousBlock = currentBlock;
        }
        if (currentBlock->visitCount > __MAXIMUM_BLOCK_VISIT) {
            currentEntry->address = 0x0;
            if (lastInvalidEntry < 0) {
                lastInvalidEntry = i;
            }
        } else {
            currentBlock->sampleCount++;
            if (lastInvalidEntry > -1) {
                entries[lastInvalidEntry].memOpId = i;
                lastInvalidEntry = -1;
            }
            if (currentBlock->visitCount == __MAXIMUM_BLOCK_VISIT) {
                currentBlock->saturationPoint = totalNumberOfAccesses;
                saturatedBlockCount++;
            }
        }
    }
    if(lastInvalidEntry > -1){
        entries[lastInvalidEntry].memOpId = i;
    }
#endif

    processSamples_EventSampling (entries, startIndex, lastIndex);

    entries->lastFreeIdx = 1;
}

float 
getPercentage (Counter_t val1,Counter_t val2)
{
    if (!val2) {
        return 0.0;
    }
    return ((100.0*val1)/val2);
}

float 
getHitPercentage (BasicBlockInfo* block, SamplingType stype, uint32_t sys, uint32_t lvl)
{
    Counter_t total = block->hitMissCounters[STATUS_IDX(sys,stype,lvl,cache_hit)] + 
                      block->hitMissCounters[STATUS_IDX(sys,stype,lvl,cache_miss)];
    return getPercentage(block->hitMissCounters[STATUS_IDX(sys,stype,lvl,cache_hit)],total);
}

void 
MetaSim_endFuncCall_Simu (char* base,uint32_t entryCount,const char* comment) 
{
    uint32_t      i;
    uint32_t      j;
    uint32_t      k;

    BufferEntry* entries = (BufferEntry*)base;
    uint32_t lastIndex = entries->lastFreeIdx;
    lastIndex--;

    fprintf(stdout,"MetaSim_endFuncCall(0x%p,%d,%s,%d)\n",base,entryCount,comment,entries->lastFreeIdx);

    MetaSim_simulFuncCall_Simu(base,entryCount,comment);

    Counter_t processedAllCount = 0;
    Counter_t processedEventCount = 0;
    Counter_t processedAddressCount = 0;
    FILE* fp = openOutputFile(comment);
    if(fp) {
        BasicBlockInfo* currentBlock = NULL;
        for (i=0 ; i<blockCount ; i++) {
            currentBlock = (blocks + i);
            if (currentBlock->visitCount[no_sampling]) {
                processedAllCount += currentBlock->visitCount[no_sampling];
                processedEventCount += currentBlock->visitCount[event_sampling];
                processedAddressCount += currentBlock->visitCount[address_sampling];
            }
        }
        fprintf(fp,"# buffer           = %d\n", entryCount);
        fprintf(fp,"# blockCount       = %d\n", blockCount);
        fprintf(fp,"# totalAll         = %lld\n", totalNumberOfAccesses[no_sampling]);
        fprintf(fp,"# totalEvent       = %lld\n", totalNumberOfAccesses[event_sampling]);
        fprintf(fp,"# totalAddress     = %lld\n", totalNumberOfAccesses[address_sampling]);
        fprintf(fp,"# processedAll     = %lld\n", processedAllCount); 
        fprintf(fp,"# processedEvent   = %lld\n", processedEventCount); 
        fprintf(fp,"# processedAddress = %lld\n", processedAddressCount); 
#ifdef USE_SATURATION
        fprintf(fp,"# saturate  = %lld\n",lastSaturationAccess); 
#endif

#ifdef NO_SAMPLING_MODE
        fprintf(fp,"# sampling  = no\n");
#else
        fprintf(fp,"# sampling  = yes\n");
#ifdef EXTENDED_SAMPLING
        fprintf(fp,"# extended-sampling  = yes (intervals are averages)\n");
#endif
        fprintf(fp,"# ignoreint = %lld\n",(uint64_t)__IGNORING_INTERVAL_MAX);
        fprintf(fp,"# sampleint = %lld\n",(uint64_t)__SAMPLING_INTERVAL_MAX);
#endif
#ifdef USE_SATURATION
        fprintf(fp,"# maxvisits = %lld\n",(uint64_t)__MAXIMUM_BLOCK_VISIT);
#endif
#ifdef PER_SET_RECENT
        fprintf(fp,"# recentptr = per set\n");
#else
        fprintf(fp,"# recentptr = per cache\n");
#endif
        fprintf(fp, "#\n");
        SamplingType stype;
        for (j=0 ; j<systemCount ; ++j) {
            MemoryHierarchy* memoryHierarchy = (systems + j);
            fprintf (fp, "# sysid%d =", memoryHierarchy->index);
            for (stype=0 ; stype<Total_SamplingType ; ++stype)
            {
               fprintf (fp, " <%s:", samplingName(stype));
               for (k=0 ; k<memoryHierarchy->levelCount ; ++k) {
                  Cache* cache = &(memoryHierarchy->levels[k]);
                  Counter_t total = cache->hitMissCounters[stype*Total_AccessStatus+cache_hit] +
                                    cache->hitMissCounters[stype*Total_AccessStatus+cache_miss];
                  fprintf (fp, " l%1d[%lld,%lld(%5.2f)]", k,
                         total,
                         cache->hitMissCounters[stype*Total_AccessStatus+cache_hit],
                         getPercentage(cache->hitMissCounters[stype*Total_AccessStatus+cache_hit], total));
               }
               fprintf (fp, ">");
            }
            fprintf (fp, "\n");
        }

        fprintf(fp, "#\n");
        fprintf (fp, "#block <seqid> [<samplingname>: <visitcount> ...]\n");
        fprintf (fp, "#\tsys <sysid> <samplingname> lvl <cachelvl> <hitcount> <miscount> <hitpercent>\n");
#ifdef USE_SATURATION
        fprintf (fp, "#\tsaturation <satpercent>\n");
#endif

        for (i=0 ; i<blockCount ; ++i) {
            currentBlock = (blocks + i);
            if (currentBlock->visitCount[no_sampling]) {
                fprintf(fp,"block\t%d", i);
                for (stype=0 ; stype<Total_SamplingType ; ++stype)
                   fprintf (fp, "\t%s: %lld", 
                       samplingName(stype), currentBlock->visitCount[stype]);
                fprintf (fp, "\n");
                
                for (j=0 ; j<systemCount ; ++j) {
                    MemoryHierarchy* memoryHierarchy = (systems + j);
                    for (stype=0 ; stype<Total_SamplingType ; ++stype)
                    {
                       for (k=0 ; k<memoryHierarchy->levelCount ; ++k) {
                           Cache* cache = &(memoryHierarchy->levels[k]);
                           Counter_t total = currentBlock->hitMissCounters[STATUS_IDX(j,stype,k,cache_hit)] + 
                                             currentBlock->hitMissCounters[STATUS_IDX(j,stype,k,cache_miss)];
                           fprintf(fp,"\tsys\t%2d\t%s\tlvl\t%1d\t%lld\t%lld\t%10.6f\n",
                               memoryHierarchy->index,samplingName(stype),k+1,
                               currentBlock->hitMissCounters[STATUS_IDX(j,stype,k,cache_hit)],
                               currentBlock->hitMissCounters[STATUS_IDX(j,stype,k,cache_miss)],
                               getHitPercentage(currentBlock,stype,j,k));
                       }
                    }
                }

#ifdef USE_SATURATION
                if(!currentBlock->saturationPoint){
                    fprintf(fp,"\tsaturation\t%5.2f\n",100.0);
                } else {
                    fprintf(fp,"\tsaturation\t%5.2f\n",getPercentage(currentBlock->saturationPoint,
                                                                   totalNumberOfAccesses));
                }
#endif
            }
        }

        fclose(fp);
    }
}
