#include <XCoffFile.h>

#include <CustomDataExtender.h>
#include <FindAddressCoverage.h>
#include <popt.h>

typedef enum {
    unknown_inst_type = 0,
    identical_inst_type,
    frequency_inst_type,
    simulation_inst_type,
    simucntr_inst_type,
    bbtrace_inst_type,
    countblocks_inst_type,
    data_extender_type,
    address_coverage_type,
    Total_InstrumentationType
} InstrumentationType;


int main (int argc, const char** argv) {

    char*    typeName  = NULL;
    char*    execName  = NULL;
    char*    inptName  = NULL;
    char*    extension = NULL;
    char*    libPath   = NULL;
    uint32_t instType  = data_extender_type;

   /* Declare command-line options for popt. */
   struct poptOption requiredopts[] = {
     { "app", 'a', POPT_ARG_STRING, &execName, 0,
       "Name of executable to instrument.", NULL },
     { "inp", 'i', POPT_ARG_STRING, &inptName, 0,
       "Name of input file whose content is added into the data section.", NULL },
     POPT_TABLEEND
   };

   struct poptOption optionalopts[] = {
     { "typ", 't', POPT_ARG_STRING, &typeName, 1,
       "Instrumentation type (extdata, addrcov) [default: extdata]", NULL },
     { "ext", 'e', POPT_ARG_STRING, &extension, 0,
       "Instrumented binary extension [default: depends on selected type]", NULL },
     { "lib", 'l', POPT_ARG_STRING, &libPath, 0,
       "Path to shared library [default: read from $PMACINST_LIB_HOME]", NULL },
     POPT_TABLEEND
   };

   struct poptOption cmdlineopts[] = {
     { NULL, '\0', POPT_ARG_INCLUDE_TABLE, requiredopts, 0, 
       "Required parameters:", NULL },
     { NULL, '\0', POPT_ARG_INCLUDE_TABLE, optionalopts, 0, 
       "Optional parameters:", NULL },
     POPT_AUTOHELP
     POPT_TABLEEND
   };

   poptContext context;
   int poptresult;

   TIMER (double t = timer());

   /* Parse the command line. */
   context = poptGetContext (NULL, argc, argv, cmdlineopts, 0);

   do
   {
      poptresult = poptGetNextOpt(context);
      if (poptresult < -1) {
        fprintf(stderr, "%s: %s\n",
            poptBadOption(context, POPT_BADOPTION_NOALIAS),
            poptStrerror(poptresult));
        exit (-1);
      }
      switch (poptresult)
      {
         case 1:
            if (!strcmp (typeName, "extdata"))
            {
               instType  = data_extender_type;
               extension = "extdata";
            } else
            if (!strcmp (typeName, "addrcov"))
            {
               instType  = address_coverage_type;
               extension = "addrcov";
            } else
            {
               fprintf (stderr, "The only supported instrumentation types are 'extdata' and 'addrcov'.\n");
               exit (-4);
            }
            break;
      }
   } while (poptresult > 0);

   if (!execName) {
       fprintf (stderr,"\nError : No executable is specified\n\n");
       poptPrintHelp (context, stdout, 0);
       exit (-2);
   }

   if (!inptName) {
       fprintf (stderr,"\nError : No input file is specified\n\n");
       poptPrintHelp (context, stdout, 0);
       exit (-3);
   }

   if ((instType <= unknown_inst_type) || 
      (instType >= Total_InstrumentationType)) {
       fprintf(stderr,"\nError : Unknown instrumentation type\n");
       poptPrintHelp (context, stdout, 0);
       exit (-4);
   }

   if (sizeof(PowerPCInstruction) != 4){
       PRINT_ERROR("Size of instruction has to be 4 bytes");
   }

   if (sizeof(HashCode) != 8){
       PRINT_ERROR("Development error; The size of hashCode has to be 8 bytes");
   }

   if(!libPath){
       libPath = getenv("PMACINST_LIB_HOME");
       if (!libPath){
           PRINT_ERROR("Error : use --lib option or define set the PMACINST_LIB_HOME variable"); 
       }
   }
   PRINT_INFOR("The instrumentation libraries will be used from %s",libPath);
   
   if (!extension)
   {
      assert (instType == data_extender_type);
      extension = "extdata";
   }

   TIMER(double t1 = timer());

   PRINT_INFOR("******** Instrumentation Beginning ********");
   XCoffFile xcoffFile(execName);

   xcoffFile.parse();
   xcoffFile.briefPrint();

   TIMER(double t2 = timer();PRINT_INFOR("___timer: Instrumentation Step I parse  : %.2f",t2-t1);t1=t2);

   DEBUG_MORE(
       xcoffFile.displaySymbols();
       xcoffFile.print();
   );

   TIMER(t2 = timer();PRINT_INFOR("___timer: Instrumentation Step II Line  : %.2f",t2-t1);t1=t2);

   DEBUG(xcoffFile.testBitSet(););

   TIMER(t2 = timer();PRINT_INFOR("___timer: Instrumentation Step III Loop : %.2f",t2-t1);t1=t2);

   XCoffFileGen* xcoffFileGen = NULL;
   if (instType == data_extender_type) {
       xcoffFileGen = new CustomDataExtender (&xcoffFile, extension, inptName);
   } else
   if (instType == address_coverage_type) {
       xcoffFileGen = new FindAddressCoverage (&xcoffFile,extension,0,inptName);
   } else {
       PRINT_ERROR("Error : unsupported instrumentation type, so quiting");
   }

   xcoffFileGen->setPathToInstLib (libPath);

   PRINT_INFOR("******** Function  Count  %d          ********",xcoffFileGen->getNumberOfAllFunctions());
   PRINT_INFOR("******** Basic Block Cou  %d          ********",xcoffFileGen->getNumberOfAllBlocks());
   PRINT_INFOR("******** Memory Op Count  %d          ********",xcoffFileGen->getNumberOfAllMemoryOps());
   PRINT_INFOR("******** FloatP Op Count  %d          ********",xcoffFileGen->getNumberOfAllFloatPOps());
   PRINT_INFOR("******** Instrument Point %d          ********",xcoffFileGen->getNumberOfInstPoints(NULL));

   xcoffFileGen->instrument();

   TIMER(t2 = timer();PRINT_INFOR("___timer: Instrumentation Step IV Instr : %.2f",t2-t1);t1=t2);

   PRINT_INFOR("******** Original File Si %d          ********",xcoffFile.getFileSize());
   PRINT_INFOR("******** Instrumented Siz %d          ********",xcoffFileGen->getInstrumentedFileSize());

   uint32_t written = xcoffFileGen->dump();
   PRINT_INFOR("******** Output File Siz  %d          ********",written);

   TIMER(t2 = timer();PRINT_INFOR("___timer: Instrumentation Step V Dump   : %.2f",t2-t1);t1=t2);

   xcoffFileGen->verify(written);

   PRINT_INFOR("******** Instrumentation Successfull ********");

   TIMER(t = timer()-t;PRINT_INFOR("___timer: Total Execution Time          : %.2f",t););

   PRINT_INFOR("\n");
   PRINT_INFOR("******** DONE ******** SUCCESS ***** SUCCESS ***** SUCCESS ********\n");
   return 0;
}
