#include <Base.h>
#include <XCoffFile.h>
#include <BinaryFile.h>
#include <SectHeader.h>
#include <Instruction.h>
#include <RawSection.h>
#include <strings.h>
#include <Generate.h>
#include <Instruction.h>

char* VersionTag::tag = "revision REVISION";

extern uint64_t nextAlignAddressHalfWord(uint64_t addr){
    addr += (addr % 2 ? 2 : 0);
    addr >>= 1;
    addr <<= 1;
    return addr;
}
extern uint64_t nextAlignAddressWord(uint64_t addr){
    addr += (addr % 4 ? 4 : 0);
    addr >>= 2;
    addr <<= 2;
    return addr;
}
extern uint64_t nextAlignAddressDouble(uint64_t addr){
    addr += (addr % 8 ? 8 : 0);
    addr >>= 3;
    addr <<= 3;
    return addr;
}

HashCode::HashCode(uint32_t s){
    entry.bits = INVALID_FIELD;
    if(validSection(s)){
        entry.fields.section  = s;
        entry.fields.function = INVALID_FIELD;
        entry.fields.block    = INVALID_FIELD;
        entry.fields.memop    = INVALID_FIELD;
    }
}

HashCode::HashCode(uint32_t s,uint32_t f){
    entry.bits = INVALID_FIELD;
    if(validSection(s) && validFunction(f)){
        entry.fields.section  = s;
        entry.fields.function = ++f;
        entry.fields.block    = INVALID_FIELD;
        entry.fields.memop    = INVALID_FIELD;
    }
}

HashCode::HashCode(uint32_t s,uint32_t f,uint32_t b){
    entry.bits = INVALID_FIELD;
    if(validSection(s) && validFunction(f) && validBlock(b)){
        entry.fields.section  = s;
        entry.fields.function = ++f;
        entry.fields.block    = ++b;
        entry.fields.memop    = INVALID_FIELD;
    }
}

HashCode::HashCode(uint32_t s,uint32_t f,uint32_t b,uint32_t m){
    entry.bits = INVALID_FIELD;
    if(validSection(s) && validFunction(f) && validBlock(b) && validMemop(m)){
        entry.fields.section  = s;
        entry.fields.function = ++f;
        entry.fields.block    = ++b;
        entry.fields.memop    = ++m;
    }
}

uint32_t BaseGen::setFileOffset(uint32_t offset){ 
    if(!sizeInBytes){
        return offset;
    }
    fileOffset = (uint32_t)nextAlignAddressHalfWord(offset); 
    return (fileOffset+sizeInBytes); 
}

#define PAGE_MASK 0xfff
uint32_t RawSectionGen::setFileOffset(uint32_t offset){

    if(!sizeInBytes){
        return offset;
    }

    if(rawSection->IS_SECT_TYPE(TEXT) || 
       rawSection->IS_SECT_TYPE(DATA))
    {
        uint64_t currFilePointer = offset;
        uint64_t prevFilePointer = rawSection->getSectHeader()->GET(s_scnptr);

        uint64_t currRemainder = PAGE_MASK & currFilePointer;
        uint64_t prevRemainder = PAGE_MASK & prevFilePointer;

        if(currRemainder > prevRemainder){
            currFilePointer += (PAGE_MASK+1);
        }
        currFilePointer = (currFilePointer & ~PAGE_MASK) | (prevFilePointer & PAGE_MASK);

        DEBUG(
            uint64_t virtualAddress = rawSection->getSectHeader()->GET(s_vaddr);
            ASSERT((virtualAddress & PAGE_MASK) == (currFilePointer & PAGE_MASK));
        );

        offset = (uint32_t)currFilePointer; 

        fileOffset = offset; 
        return (fileOffset+sizeInBytes); 
    } 

    return BaseGen::setFileOffset(offset);
}

void BaseGen::initInstrumentationBuffer(XCoffFileGen* xCoffGen){
    Base* parsedBase = getParsedBase();
    uint32_t bufferSize = parsedBase->getInstrumentationSize(xCoffGen);
    if(!bufferSize){
        return;
    }
    sizeInBytes = bufferSize;
    instBuffer = new char[sizeInBytes];
    bzero(instBuffer,sizeInBytes);
}

void BaseGen::instrument(XCoffFileGen* xCoffGen){
    Base* parsedBase = getParsedBase();

    uint32_t bufferSize = parsedBase->getInstrumentationSize(xCoffGen);
    ASSERT((!bufferSize || (sizeInBytes == bufferSize)) && 
           "FATAL : The buffer size is not same as instrumentation");
    ASSERT((!bufferSize || !hasInvalidFileOffset()) && 
           "FATAL : The buffer size is not same as instrumentation");

    parsedBase->instrument(instBuffer,xCoffGen,this);
}

void BaseGen::dump(BinaryOutputFile* bf){
    if(sizeInBytes){
        ASSERT(instBuffer && "FATAL : the non-zero size segment but the buffer is not allocated");
        bf->copyBytes(instBuffer,sizeInBytes,fileOffset);
    }
}

void BaseGen::printOffsetMapping(){
    Base* parsedBase = getParsedBase();
    PRINT_INFOR("%30s OFFSET (%d -- %d) SIZE(%d -- %d)",
        parsedBase->briefName(),
        getFileOffset(),parsedBase->getFileOffset(),
        getSizeInBytes(),parsedBase->getSizeInBytes());

}

void BaseGen::verifyIdentical(char* origFileName,char* instFileName){

    FILE* origFile = NULL;
    FILE* instFile = NULL;

    origFile = fopen(origFileName,"r");
    instFile = fopen(instFileName,"r");

    if(!origFile || !instFile){
        PRINT_ERROR("One of the original and instrumented file can not be opened");
        return;
    }

    Base* parsedBase = getParsedBase();
    ASSERT((getFileOffset() == parsedBase->getFileOffset()) && 
           "FATAL : Idemtical inst should have same offsets");

    if(hasInvalidFileOffset())
        return;

    ASSERT(getSizeInBytes() == parsedBase->getSizeInBytes());

    uint32_t startOffset = getFileOffset();
    uint32_t endOffset = startOffset + getSizeInBytes();

    int32_t check1 = fseek(origFile,startOffset,SEEK_SET);
    int32_t check2 = fseek(instFile,startOffset,SEEK_SET);
    if(check1 || check2){
        PRINT_ERROR("Original or Instrumented file can not be seeked");
    }

    for(;startOffset<endOffset;startOffset += sizeof(uint32_t)){
        bool fatalDiff = false;

        uint32_t word1 = 0x0;
        uint32_t word2 = 0x0;
        char* ptr1 = (char*)&word1;
        char* ptr2 = (char*)&word2;

        check1 = fread(ptr1,sizeof(uint32_t),1,origFile);
        check2 = fread(ptr2,sizeof(uint32_t),1,instFile);

        if((check1 || check2) && (((uint32_t)check1 != 1) || ((uint32_t)check2 != 1))){
            PRINT_ERROR("Original or Instrumented file can not be read");
        }

        uint32_t remainder = endOffset-startOffset;
        if(remainder < sizeof(uint32_t)){
            for(uint32_t j=0;j<remainder;j++,ptr1++,ptr2++){
                fatalDiff = fatalDiff || (*ptr1 != *ptr2);
            }
        } else {
            fatalDiff = (word1 != word2);
        }

        if(fatalDiff){
            PRINT_DEBUG("---- %24s --> %9d -- %#10x %#10x %d",
                parsedBase->briefName(),startOffset,word1,word2,
                (remainder < sizeof(uint32_t) ? remainder : sizeof(uint32_t)));
            ASSERT(0 && "FATAL: the instrumented and original files differ even it is identical");
        }
    }

    fclose(origFile);
    fclose(instFile);
}

bool Base::includesFileOffset(uint32_t offset){
    if(hasInvalidFileOffset())
        return false;
    uint32_t begin = getFileOffset();
    uint32_t end = begin + getSizeInBytes();
    return ((begin <= offset) && (offset < end));
}

bool BaseGen::includesFileOffset(uint32_t offset){
    if(hasInvalidFileOffset())
        return false;
    uint32_t begin = getFileOffset();
    uint32_t end = begin + getSizeInBytes();
    return ((begin <= offset) && (offset < end));
}

uint32_t BaseGen::convertFileOffset(uint32_t oldOffset){
     Base* parsedBase = getParsedBase();
     if(!parsedBase->includesFileOffset(oldOffset))
        return oldOffset;
     uint32_t diff = oldOffset - parsedBase->getFileOffset();
     uint32_t newOffset = getFileOffset() + diff;
     PRINT_DEBUG("Converting file offsets from %9d ----> %9d",oldOffset,newOffset)
     return newOffset;
}

void BaseGen::writeHalfWord(uint32_t offset,uint16_t value){
    if(offset >= sizeInBytes){
        return;
	}
    if((offset+sizeof(uint16_t)) > sizeInBytes){
        return;
	}
    if(!instBuffer)
        return;
    memcpy(instBuffer+offset,&value,sizeof(uint16_t));
}
void BaseGen::writeWord(uint32_t offset,uint32_t value){
    if(offset >= sizeInBytes)
        return;
    if((offset+sizeof(uint32_t)) > sizeInBytes)
        return;
    if(!instBuffer)
        return;
    memcpy(instBuffer+offset,&value,sizeof(uint32_t));
}
void BaseGen::writeDouble(uint32_t offset,uint64_t value){
    if(offset >= sizeInBytes)
        return;
    if((offset+sizeof(uint64_t)) > sizeInBytes)
        return;
    if(!instBuffer)
        return;
    memcpy(instBuffer+offset,&value,sizeof(uint64_t));
}
void BaseGen::writeInstructions(uint32_t offset,uint32_t insnCount,Instruction* insnBuffer){
    for (uint32_t i = 0; i < insnCount; i++, offset += sizeof(uint32_t)){
        writeWord(offset,insnBuffer[i].bits());
    }
    DEBUG_MORE(
        PRINT_INFOR("Total number of Instructions is %d",insnCount);
        for (uint32_t i = 0; i < insnCount; i++){
            insnBuffer[i].print(offset + (sizeof(uint32_t)*i),true);
        }
    );
}
void BaseGen::writeString(uint32_t offset,char* buff,uint32_t len){
    if (offset >= sizeInBytes)
        return;
    if ((offset+len) > sizeInBytes)
        return;
    if (!instBuffer)
        return;
    memcpy(instBuffer+offset,buff,len);
}

#include <time.h>
#include <sys/time.h>
#include <sys/timers.h>

double timer(){
	struct timespec time_record;;
	getclock(TIMEOFDAY,&time_record);
	return ((double)time_record.tv_sec+((double)time_record.tv_nsec) * 1.0e-9);

}


