#include <RelocationTable.h>
#include <LineInfoTable.h>
#include <SectHeader.h>
#include <BinaryFile.h>

#define LINE_RELOC_OVERFLOW 0xffff

void SectHeader::initFilePointers(BinaryInputFile* binaryInputFile){
    if((GET(s_size) != 0) && !IS_SECT_TYPE(OVRFLO) && !IS_SECT_TYPE(BSS)){
        rawDataPtr = binaryInputFile->fileOffsetToPointer(GET(s_scnptr));
    }
    if((GET(s_nreloc) != 0) && !IS_SECT_TYPE(OVRFLO)){
        numOfRelocations = (uint32_t)GET(s_nreloc);
        if(numOfRelocations){
            relocationPtr = binaryInputFile->fileOffsetToPointer(GET(s_relptr));
            ASSERT((IS_SECT_TYPE(DATA) || IS_SECT_TYPE(TEXT)) && 
                "FATAL : Only text and Data can have reloc");
        }
    }
    if((GET(s_nlnno) != 0) && !IS_SECT_TYPE(OVRFLO)){
        numOfLineInfo = (uint32_t)GET(s_nlnno);
        if(numOfLineInfo){
            ASSERT(IS_SECT_TYPE(TEXT) && 
                "FATAL : Only text and Data can have reloc");
            lineInfoPointer = binaryInputFile->fileOffsetToPointer(GET(s_lnnoptr));
        }
    }
}

bool SectHeader::inRange(uint64_t address){ 
    if(!GET(s_vaddr) || IS_SECT_TYPE(OVRFLO)) 
        return false; 
    return (((GET(s_vaddr)) <= address) && (address < (GET(s_vaddr) + GET(s_size)))); 
}

RelocationTable* SectHeader::readRelocTable(BinaryInputFile* binaryInputFile,XCoffFile* xcoff){
    relocationTable = NULL;
    if(!numOfRelocations || !relocationPtr)
        return NULL;

    relocationTable = new RelocationTable(relocationPtr,numOfRelocations,xcoff);
    relocationTable->read(binaryInputFile);
    return relocationTable;
}

LineInfoTable* SectHeader::readLineInfoTable(BinaryInputFile* binaryInputFile,XCoffFile* xcoff){
    lineInfoTable = NULL;
    if(!numOfLineInfo || !lineInfoPointer)
        return NULL;

    lineInfoTable = new LineInfoTable(lineInfoPointer,numOfLineInfo,xcoff);
    lineInfoTable->read(binaryInputFile);
    return lineInfoTable;
}

bool SectHeader::verify() {
    if(IS_SECT_TYPE(BSS)){
        ASSERT((GET(s_scnptr) == 0) && "FATAL : BSS section does not have any raw data\n");
    }
    if(!IS_SECT_TYPE(DATA) && !IS_SECT_TYPE(TEXT) && !IS_SECT_TYPE(OVRFLO)){
        ASSERT(!GET(s_relptr) && "FATAL : Only TEXT, DATA, OVERFLO sections can have relocation data\n");
    } 
    if(!IS_SECT_TYPE(TEXT) && !IS_SECT_TYPE(OVRFLO)){
        ASSERT(!GET(s_lnnoptr) && "FATAL : Only TEXT section can have line info\n");
    }
    if(GET(s_nreloc) == LINE_RELOC_OVERFLOW){
        ASSERT((GET(s_nreloc) == GET(s_nlnno)) && 
                "FATAL : If overflow the #reloc and #lineno should match\n");
        ASSERT((IS_SECT_TYPE(DATA) || IS_SECT_TYPE(TEXT)) && 
                "FATAL :Only DAT and TEXT can have overflow\n");
    }
    if(IS_SECT_TYPE(OVRFLO)){
        ASSERT((GET(s_nreloc) == GET(s_nlnno)) && 
                "FATAL : The section number in overflow should match\n");
        ASSERT(!GET(s_size) && !GET(s_scnptr) && 
                "FATAL : other field in overflow should be 0\n");
    }

    return true;
}

void SectHeader::setOverFlowSection(SectHeader* sh) { 
    ASSERT((IS_SECT_TYPE(DATA) || IS_SECT_TYPE(TEXT)) &&
        "FATAL : Only text and Data can have reloc");

    overFlowSection = sh; 
    numOfRelocations = (uint32_t)(sh->GET(s_paddr));
    numOfLineInfo = (uint32_t)(sh->GET(s_vaddr));
}

void SectHeader::print() { 
    char sname[__MAX_STRING_SIZE];

    memcpy(sname,GET(s_name),8);
    sname[8] = '\0';

    PRINT_INFOR("SECTHEADER %s",getTypeName());
    PRINT_INFOR("\tName  : %s",sname);
    PRINT_INFOR("\tType  : %#x [OVRFLO TYPCHK DEBUG LOADER] [R R INFO EXCEPT] [BSS DATA TEXT R] [PAD R R R]",GET(s_flags));
    PRINT_INFOR("\tPVads : %#llx %#llx",GET(s_paddr),GET(s_vaddr));
    PRINT_INFOR("\tRaw   : %#llx %lld bytes",GET(s_scnptr),GET(s_size));
    PRINT_INFOR("\tReloc : %#llx %d",GET(s_relptr),GET(s_nreloc));
    PRINT_INFOR("\tLines : %#llx %d",GET(s_lnnoptr),GET(s_nlnno));
}

uint32_t SectHeader32::read(BinaryInputFile* binaryInputFile){
    setFileOffset(binaryInputFile->currentOffset());

    if(!binaryInputFile->copyBytesIterate(&entry,Size__32_bit_Section_Header)){
        PRINT_ERROR("Section header (32) can not be read");
    }
    verify();
    initFilePointers(binaryInputFile);
    
    return Size__32_bit_Section_Header;
}

uint32_t SectHeader64::read(BinaryInputFile* binaryInputFile){
    setFileOffset(binaryInputFile->currentOffset());

    if(!binaryInputFile->copyBytesIterate(&entry,Size__64_bit_Section_Header)){
        PRINT_ERROR("Section header (64) can not be read");
    }
    verify();
    initFilePointers(binaryInputFile);

    return Size__64_bit_Section_Header;
}

#define CASE_SECTION_TYPE(__str)  case STYP_ ## __str: return #__str
const char* SectHeader::getTypeName(){
    switch(GET(s_flags)){
        CASE_SECTION_TYPE(TEXT);
        CASE_SECTION_TYPE(DATA);
        CASE_SECTION_TYPE(BSS);
        CASE_SECTION_TYPE(PAD);
        CASE_SECTION_TYPE(LOADER);
        CASE_SECTION_TYPE(DEBUG);
        CASE_SECTION_TYPE(TYPCHK);
        CASE_SECTION_TYPE(EXCEPT);
        CASE_SECTION_TYPE(OVRFLO);
        CASE_SECTION_TYPE(INFO);
        default: {
             ASSERT(NULL && "FATAL : Ths symbol type is incorrect");
             return NULL;

        }
    }
}

