#include <Instruction.h>
#include <DemangleWrapper.h>
#include <SymbolTable.h>
#include <StringTable.h>
#include <SectHeader.h>
#include <RawSection.h>
#include <Auxilary.h>
#include <XCoffFile.h>
#include <BinaryFile.h>

int compareSymbolValue(const void* arg1,const void* arg2){
    Symbol* sym1 = *((Symbol**)arg1);
    Symbol* sym2 = *((Symbol**)arg2);
    uint64_t vl1 = sym1->GET(n_value);
    uint64_t vl2 = sym2->GET(n_value);

    if(vl1 < vl2)
        return -1;
    if(vl1 > vl2)
        return 1;

    ASSERT((sym1->GET(n_scnum) == sym2->GET(n_scnum)) && 
           "FATAL: Two symbols with the same address but different sections");

    return 0;
}

int searchSymbolValue(const void* arg1,const void* arg2){
    uint64_t key = *((uint64_t*)arg1);
    Symbol* sym = *((Symbol**)arg2);
    uint64_t val = sym->GET(n_value);

    if(key < val)
        return -1;
    if(key > val)
        return 1;
    return 0;
}

bool Symbol::unknownSymbol(){
    if(isAuxilary())
        return false;
    if(IS_SYMB_TYPE(ALIAS) || IS_SYMB_TYPE(ARG) || IS_SYMB_TYPE(AUTO) || 
       IS_SYMB_TYPE(EFCN) || IS_SYMB_TYPE(ENTAG) || IS_SYMB_TYPE(EOS) || 
       IS_SYMB_TYPE(EXTDEF) || IS_SYMB_TYPE(FIELD) || IS_SYMB_TYPE(HIDDEN) || 
       IS_SYMB_TYPE(LABEL) || IS_SYMB_TYPE(LINE) || IS_SYMB_TYPE(MOE) || 
       IS_SYMB_TYPE(MOS) || IS_SYMB_TYPE(MOU) || IS_SYMB_TYPE(REG) || 
       IS_SYMB_TYPE(REGPARM) || IS_SYMB_TYPE(STRTAG) || IS_SYMB_TYPE(TPDEF) || 
       IS_SYMB_TYPE(ULABEL) || IS_SYMB_TYPE(UNTAG) || IS_SYMB_TYPE(USTATIC))
       return true;
    return false;
}

bool Symbol::stringInDebugSection(){
    if(isAuxilary())
        return false;

    if(IS_SYMB_TYPE(BCOMM) || IS_SYMB_TYPE(DECL) || IS_SYMB_TYPE(ECOML) || 
       IS_SYMB_TYPE(ECOMM) || IS_SYMB_TYPE(ENTRY) || IS_SYMB_TYPE(FUN) || 
       IS_SYMB_TYPE(GSYM) || IS_SYMB_TYPE(LSYM) || IS_SYMB_TYPE(PSYM) || 
       IS_SYMB_TYPE(RPSYM) || IS_SYMB_TYPE(RSYM) || IS_SYMB_TYPE(STSYM) ||
       IS_SYMB_TYPE(TCSYM))
        return true;
    return false;
}


char* Symbol::getName(StringTable* stringTable,DebugSection* debugRawSect){
    char* ptr = NULL;
    if(!stringInDebugSection()){
        ASSERT(stringTable && "FATAL : String is needed from string table");
        ptr = getNameStringTable(stringTable);
    } else {
        ASSERT(debugRawSect && "FATAL : String is needed from debug");
        ptr = getNameDebugSection(debugRawSect);
    }
    ASSERT(ptr && "FATAL : the symbol name does not exist");
    return ptr;
}

void Symbol::print(StringTable* stringTable,DebugSection* debugRawSect,bool followAux){

    char* ptr = getName(stringTable,debugRawSect);

    DemangleWrapper wrapper;
    char* demangled = wrapper.demangle_combined(ptr);

    PRINT_INFOR("SYM\t%10s [%7d](sec %2d)(fl 0x00)(ty %3x)(scl %3d) (nx %d) 0x%016llx %s --- %s",
            getTypeName(),getIndex(),GET(n_scnum),GET(n_type),GET(n_sclass),
            GET(n_numaux),GET(n_value),ptr,demangled);
    free(ptr);

    if(followAux && getNext()){
        for(SymbolBase* s = getNext(); s; s = s->getNext()){
            s->print(stringTable,debugRawSect,followAux);
        }
    }
}

#define NO_NAME_OFFSET (uint32_t)(-1)
uint32_t Symbol32::getNameOffset(){
    uint32_t offset = NO_NAME_OFFSET;
    uint32_t zeroes = GET(n_zeroes);
    if(!zeroes){
        offset = GET(n_offset);
    } 
    return offset;
}

uint32_t Symbol64::getNameOffset(){
    uint32_t offset = GET(n_offset);
    return offset;
}

char* Symbol32::getNameDebugSection(DebugSection* rawDebugSect){
    char* ret = NULL;
    char tmpBuffer[9];
    if(!rawDebugSect)
        return ret;

    uint32_t offset = getNameOffset();
    if(offset == NO_NAME_OFFSET){
        strncpy(tmpBuffer,GET(n_name),8);
        tmpBuffer[8] = '\0';
        ret = tmpBuffer;
    } else {
        ret = rawDebugSect->getString(offset);
    }
    return strdup(ret);
}

char* Symbol64::getNameDebugSection(DebugSection* rawDebugSect){
    char* ret = NULL;
    if(!rawDebugSect)
        return ret;

    uint32_t offset = getNameOffset();
    ret = rawDebugSect->getString(offset);
    return strdup(ret);
}

char* Symbol32::getNameStringTable(StringTable* stringTable){

    char* ret = NULL;
    char tmpBuffer[9];

    if(!stringTable)
        return ret;

    uint32_t offset = getNameOffset();
    if(offset == NO_NAME_OFFSET){
        strncpy(tmpBuffer,GET(n_name),8);
        tmpBuffer[8] = '\0';
        ret = tmpBuffer;
    } else {
        ret = stringTable->getString(offset);
    }
    return strdup(ret);
}
char* Symbol64::getNameStringTable(StringTable* stringTable){
    char* ret = NULL;
    if(!stringTable)
        return ret;

    uint32_t offset = getNameOffset();
    ret = stringTable->getString(offset);
    return strdup(ret);
}

void Symbol32::changeValueCopy(uint64_t value,char* buff){
    SYMENT newEntry;
    newEntry = entry;
    newEntry.n_value = value;
    memcpy(buff,&newEntry,Size__NN_bit_SymbolTable_Entry);
}

void Symbol64::changeValueCopy(uint64_t value,char* buff){
    SYMENT_64 newEntry;
    newEntry = entry;
    newEntry.n_value = value;
    memcpy(buff,&newEntry,Size__NN_bit_SymbolTable_Entry);
}


Symbol* binarySearch(Symbol** symbols,uint32_t symbolCount,uint64_t value){

    if(!symbolCount)
        return NULL;

    Symbol* ret = NULL;

    uint32_t beginIndex = 0;
    uint32_t endIndex = symbolCount;
    
    uint32_t midIndex = 0;
    Symbol* midSymbol = NULL;
    uint64_t midVal = 0;

    while(beginIndex < endIndex){
        midIndex = (beginIndex+endIndex) / 2;
        midSymbol = symbols[midIndex];
        midVal = midSymbol->GET(n_value);

        if(midVal > value){
            endIndex = midIndex;
        } else if(midVal < value){
            beginIndex = midIndex+1;
        } else {
            ret = midSymbol;
            break;
        }
    }

    if(!ret){
        if(value < midVal){
            if(midIndex) {
                midIndex--;
                ret = symbols[midIndex];
            }
        } else {
            if(midIndex < (symbolCount-1)){
                ret = midSymbol;
            }
        }
    }

    return ret;
}

Symbol* Symbol::findSymbol(Symbol** symbols,uint32_t symbolCount,uint64_t value){
    return binarySearch(symbols,symbolCount,value);
}



#define CASE_SYMBOL_TYPE(__str)  case C_ ## __str: return #__str
const char* Symbol::getTypeName(){
    switch(GET(n_sclass)){
        CASE_SYMBOL_TYPE(ALIAS);
        CASE_SYMBOL_TYPE(ARG);
        CASE_SYMBOL_TYPE(AUTO);
        CASE_SYMBOL_TYPE(BCOMM);
        CASE_SYMBOL_TYPE(BINCL);
        CASE_SYMBOL_TYPE(BLOCK);
        CASE_SYMBOL_TYPE(BSTAT);
        CASE_SYMBOL_TYPE(DECL);
        CASE_SYMBOL_TYPE(ECOML);
        CASE_SYMBOL_TYPE(ECOMM);
        CASE_SYMBOL_TYPE(EFCN);
        CASE_SYMBOL_TYPE(EINCL);
        CASE_SYMBOL_TYPE(ENTAG);
        CASE_SYMBOL_TYPE(ENTRY);
        CASE_SYMBOL_TYPE(EOS);
        CASE_SYMBOL_TYPE(ESTAT);
        CASE_SYMBOL_TYPE(EXT);
        CASE_SYMBOL_TYPE(EXTDEF);
        CASE_SYMBOL_TYPE(FCN);
        CASE_SYMBOL_TYPE(FIELD);
        CASE_SYMBOL_TYPE(FILE);
        CASE_SYMBOL_TYPE(FUN);
        CASE_SYMBOL_TYPE(GSYM);
        CASE_SYMBOL_TYPE(HIDDEN);
        CASE_SYMBOL_TYPE(HIDEXT);
        CASE_SYMBOL_TYPE(INFO);
        CASE_SYMBOL_TYPE(LABEL);
        CASE_SYMBOL_TYPE(LINE);
        CASE_SYMBOL_TYPE(LSYM);
        CASE_SYMBOL_TYPE(MOE);
        CASE_SYMBOL_TYPE(MOS);
        CASE_SYMBOL_TYPE(MOU);
        CASE_SYMBOL_TYPE(NULL);
        CASE_SYMBOL_TYPE(PSYM);
        CASE_SYMBOL_TYPE(REG);
        CASE_SYMBOL_TYPE(REGPARM);
        CASE_SYMBOL_TYPE(RPSYM);
        CASE_SYMBOL_TYPE(RSYM);
        CASE_SYMBOL_TYPE(STAT);
        CASE_SYMBOL_TYPE(STRTAG);
        CASE_SYMBOL_TYPE(STSYM);
        CASE_SYMBOL_TYPE(TCSYM);
        CASE_SYMBOL_TYPE(TPDEF);
        CASE_SYMBOL_TYPE(ULABEL);
        CASE_SYMBOL_TYPE(UNTAG);
        CASE_SYMBOL_TYPE(USTATIC);
        CASE_SYMBOL_TYPE(WEAKEXT);
        default: {
            PRINT_INFOR("Some symbol that is not known - %d",GET(n_sclass));
            return "SYM_UNK_NAME";
        }
    }
}

SymbolTable::SymbolTable(char* ptr,uint32_t nsyms,XCoffFile* xcoff) 
    : Base(XCoffClassTypes_symbol_table),symbolTablePtr(ptr),
      numberOfSymbols(nsyms),stringTable(NULL),debugRawSect(NULL),
      xCoffFile(xcoff)
{
    symbols = new SymbolBase*[numberOfSymbols];
}

SymbolTable::~SymbolTable(){
    for(uint32_t i=0;i<numberOfSymbols;i++)
        delete symbols[i];
    delete[] symbols;
}

void SymbolTable::print() { 

    PRINT_INFOR("SYMBOLTABLE");
    if(symbolTablePtr && sizeInBytes && numberOfSymbols){
        PRINT_INFOR("\tSize  : %d",sizeInBytes);
        PRINT_INFOR("\tCoun  : %d",numberOfSymbols);
    } else {
        PRINT_INFOR("\tSize  : 0 EMPTY");
    }

    PRINT_INFOR("\tSyms  :");
    for(uint32_t i=0;i<numberOfSymbols;i++){
        if(!symbols[i]->isAuxilary()){
            symbols[i]->print(stringTable,debugRawSect,true);
        }
    }
}

void SymbolTable::printSymbol(uint32_t index) {
    ASSERT((index < numberOfSymbols) && "FATAL : SymbolBase does not exists");
    printSymbol(symbols[index]);
}

void SymbolTable::printSymbol(SymbolBase* sym) {
    sym->print(stringTable,debugRawSect,true);
}


SymbolBase* SymbolTable::newSymbol(SymbolBase* lastSymbol,
                                   uint8_t remaningAux,
                                   BinaryInputFile* binaryInputFile,uint32_t index,
                                   XCoffFile* xcoff)
{

    SymbolBase* ret = NULL;
    if(lastSymbol){
        ASSERT(!lastSymbol->isAuxilary() && "FATAL : last symbol should be non-axilary");
    }

    Symbol* symbol = (Symbol*)lastSymbol;
    bool delayed = false;

    if(!remaningAux){
        if(xcoff->is64Bit()){
            ret = new Symbol64(index);
        } else {
            ret = new Symbol32(index);
        }
    } else {
        if(symbol->IS_SYMB_TYPE(FILE)){
            if(xcoff->is64Bit()){
                ret = new AuxilaryFile64(index);
            } else {
                ret = new AuxilaryFile32(index);
            }
        } else if(symbol->IS_SYMB_TYPE(STAT)){ 
            ret = new AuxilarySection(index);
        } else if(symbol->IS_SYMB_TYPE(BLOCK) ||
                  symbol->IS_SYMB_TYPE(FCN)){
            if(xcoff->is64Bit()){
                ret = new AuxilaryBlock64(index);
            } else {
                ret = new AuxilaryBlock32(index);
            }
        } else if(symbol->IS_SYMB_TYPE(EXT) ||
                  symbol->IS_SYMB_TYPE(HIDEXT) ||
                  symbol->IS_SYMB_TYPE(WEAKEXT)){
            if(xcoff->is64Bit()){
                if(remaningAux == 1){
                    ret = new AuxilaryCSect64(index);
                } else {
                    delayed = true;
                    ret = new AuxilaryFunction64(index);
                }
            } else {
                if(remaningAux == 1){
                    ret = new AuxilaryCSect32(index);
                } else {
                    ret = new AuxilaryFunction32(index);
                }
            }
        } else {
            ASSERT(0 && "FATAL : there is an auxliary for an unknown type");
        }
    }

    binaryInputFile->copyBytesIterate(ret->charStream(),Size__NN_bit_SymbolTable_Entry);

    if(remaningAux){
        if((symbol->IS_SYMB_TYPE(EXT) ||
            symbol->IS_SYMB_TYPE(HIDEXT) ||
            symbol->IS_SYMB_TYPE(WEAKEXT)) &&
            (remaningAux == 1)){
                ASSERT((((Auxilary*)ret)->getAuxilaryType() == Type__Auxilary_Symbol_CSect) &&
                       "FATAL : The last aux entry should be of type CSECT");
        }
        if(delayed){
            if(((Auxilary*)ret)->getAuxilaryType() == Type__Auxilary_Symbol_Exception){
                SymbolBase* newRet = new AuxilaryException(ret->getIndex());
                memcpy(newRet->charStream(),ret->charStream(),Size__NN_bit_SymbolTable_Entry);
                delete ret;
                ret = newRet;
            }
        }
    }
    return ret;
}

uint32_t SymbolTable::read(BinaryInputFile* binaryInputFile){

    sizeInBytes = Size__NN_bit_SymbolTable_Entry * numberOfSymbols;

    uint32_t ret = 0;

    binaryInputFile->setInPointer(symbolTablePtr);
    setFileOffset(binaryInputFile->currentOffset());

    uint8_t isAux = 0;
    SymbolBase* currentSym = NULL;
    SymbolBase* prevEntry = NULL;
    for(uint32_t i = 0;i<numberOfSymbols;i++){
        symbols[i] = newSymbol(currentSym,isAux,binaryInputFile,i,xCoffFile);
        ret += Size__NN_bit_SymbolTable_Entry;

        if(isAux){
            ASSERT(currentSym && "A sym should exists before AUX");
            ASSERT(prevEntry && "Prev should exists before AUX");
            symbols[i]->setPrev(currentSym);
            prevEntry->setNext(symbols[i]);
            isAux--;
        } else {
            isAux = ((Symbol*)symbols[i])->GET(n_numaux);
            currentSym = symbols[i];
            if(((Symbol*)currentSym)->unknownSymbol()){
                PRINT_DEBUG("Information about %s is not known very well",
                            currentSym->getTypeName());
            }
        }
        prevEntry = symbols[i];
    }

    ASSERT((sizeInBytes == ret) && "FATAL : Somehow the number of read does not match");

    PRINT_INFOR("SymbolBase table contains %d symbols (%d bytes) at %#x",
                numberOfSymbols,sizeInBytes,(uint32_t)(symbolTablePtr-binaryInputFile->inPtrBase()));
    
    return ret;
}

#define INVALID_SYMBOL_LENGTH (uint64_t)-1

uint8_t SymbolTable::getStorageMapping(SymbolBase* symBase){
    if(symBase->isAuxilary())
        return (uint8_t)-1;

    Symbol* sym = (Symbol*)symBase;
    if(!sym->IS_SYMB_TYPE(EXT) &&
       !sym->IS_SYMB_TYPE(HIDEXT) &&
       !sym->IS_SYMB_TYPE(WEAKEXT))
        return (uint8_t)-1;

    Auxilary* aux = (Auxilary*)(sym->getNext());
    for(;aux;aux=(Auxilary*)aux->getNext()){
        if(aux->getAuxilaryType() == Type__Auxilary_Symbol_CSect){
            AuxilaryCSect* auxCsect = (AuxilaryCSect*)aux;
            return auxCsect->GET_A(x_smclas,x_csect);
        }
    }

    return (uint8_t)-1;
}

uint64_t SymbolTable::getSymbolLength(SymbolBase* symBase){

    if(symBase->isAuxilary())
        return 0;

    Symbol* sym = (Symbol*)symBase;
    if(!sym->IS_SYMB_TYPE(EXT) &&
       !sym->IS_SYMB_TYPE(HIDEXT) &&
       !sym->IS_SYMB_TYPE(WEAKEXT))
       return 0;

    uint8_t numaux = sym->GET(n_numaux);
    Auxilary* aux = (Auxilary*)(sym->getNext());

    if(!numaux || !aux)
        return 0;

    uint64_t ret = 0;
    AuxilaryCSect* auxCsect = NULL;

    ret = INVALID_SYMBOL_LENGTH;
    for(;aux;aux=(Auxilary*)aux->getNext()){

        uint64_t fsize = 0;
        switch(aux->getAuxilaryType()){
            case Type__Auxilary_Symbol_Function:
                fsize = ((AuxilaryFunction*)aux)->GET_A(x_fsize,x_fcn);
                break;
            case Type__Auxilary_Symbol_Exception: 
                fsize = ((AuxilaryException*)aux)->GET_A(x_fsize,x_except);
                break;
            case Type__Auxilary_Symbol_CSect: 
                auxCsect = (AuxilaryCSect*)aux;
                break;
            default: 
                break;
        }
        if(fsize && (ret == INVALID_SYMBOL_LENGTH)){
            ret = fsize; /* the first non-positive entry **/
        }
    }

    if(ret == INVALID_SYMBOL_LENGTH){
        ASSERT(auxCsect && "FATAL : the last entry in ...EXT should be a csect");

        ret = auxCsect->getLength();
        uint8_t smtype = auxCsect->GET_A(x_smtyp,x_csect);
        uint8_t lengthMod = smtype & 0x7;

        switch(lengthMod){
            case XTY_LD:
                ret = getSymbolLength(getSymbol(ret));
                break;
            case XTY_ER:
                ret = 0;
                break;
            case XTY_SD:
            case XTY_CM:
            default:
                break;
        }
    }

    return ret;
}

uint32_t SymbolTable::filterSortAddressSymbols(Symbol** symbolArray,uint32_t arrayLength){
    uint32_t currIndex = 0;
    for(uint32_t i=0;i<numberOfSymbols;i++){
        SymbolBase* base = symbols[i];
        if(base->isAuxilary())
            continue;

        Symbol* sym= (Symbol*)base;
        if(!sym->IS_SYMB_TYPE(EXT) &&
           !sym->IS_SYMB_TYPE(HIDEXT) &&
           !sym->IS_SYMB_TYPE(WEAKEXT) &&
           !sym->IS_SYMB_TYPE(BLOCK) &&
           !sym->IS_SYMB_TYPE(STAT))
            continue;

        if(!sym->GET(n_value) || (sym->GET(n_scnum) <= 0))
            continue;

        char* ptr = getSymbolName(i);
        while(*ptr == '.')
            ptr++;

        if(!strlen(ptr) ||
           !strcmp(ptr,"text") ||
           !strcmp(ptr,"data") ||
           !strcmp(ptr,"bss"))
                continue;

        if(currIndex < arrayLength)
            symbolArray[currIndex++] = sym;
        else {
            free(ptr);
            break;
        }
    }
    if(currIndex){
        qsort(symbolArray,currIndex,sizeof(Symbol*),compareSymbolValue);
    }
    return currIndex;
}
bool Symbol::builtinSaveRestore(char* ptr){

    while(*ptr == '.')
        ptr++;

    if(!strlen(ptr) ||
#if 1
       !strcmp(ptr,"fres") ||
       !strcmp(ptr,"fsav") ||
#endif
       !strncmp(ptr,"_savef",6) ||
       !strncmp(ptr,"_restf",6) ||
       !strncmp(ptr,"_savegpr",8) ||
       !strncmp(ptr,"_restgpr",8) ||
       !strncmp(ptr,"$SAVEF",6) ||
       !strncmp(ptr,"$RESTF",6) ||
       !strncmp(ptr,"Ssavef",6) ||
       !strncmp(ptr,"Srestf",6) ||
       !strcmp(ptr,"text") ||
       !strcmp(ptr,"data") ||
       !strcmp(ptr,"bss"))
        return true;

    return false;
}

uint32_t SymbolTable::filterSortBSSSymbols(Symbol** symbolArray,uint32_t arrayLength){
	SectHeader* header = getXCoffFile()->getBSSSection()->getSectHeader();
    uint16_t sectionIndex = header->getIndex();

    uint32_t currIndex = 0;
    for(uint32_t i=0;i<numberOfSymbols;i++){
        SymbolBase* base = symbols[i];

        if(base->isAuxilary())
            continue;

        Symbol* sym= (Symbol*)base;
        if(sym->GET(n_scnum) != sectionIndex)
            continue;

        if(currIndex < arrayLength)
            symbolArray[currIndex++] = sym;
        else{
            break;
        }
    }
    if(currIndex){
        qsort(symbolArray,currIndex,sizeof(Symbol*),compareSymbolValue);
    }
    return currIndex;
}

uint32_t SymbolTable::filterSortFuncSymbols(Symbol** symbolArray,uint32_t arrayLength,SectHeader* header){

    uint16_t sectionIndex = header->getIndex();

    uint32_t currIndex = 0;
    for(uint32_t i=0;i<numberOfSymbols;i++){
        SymbolBase* base = symbols[i];

        if(base->isAuxilary())
            continue;

        Symbol* sym= (Symbol*)base;
        if(sym->GET(n_scnum) != sectionIndex)
            continue;

        if(!sym->IS_SYMB_TYPE(EXT) &&
           !sym->IS_SYMB_TYPE(HIDEXT) &&
           !sym->IS_SYMB_TYPE(WEAKEXT))
            continue;

        if(!header->inRange(sym->GET(n_value)))
            continue;

        char* ptr = getSymbolName(i);
        if(!Symbol::builtinSaveRestore(ptr)){
            if(currIndex < arrayLength)
                symbolArray[currIndex++] = sym;
            else{
                free(ptr);
                break;
            }
        }

        free(ptr);
    }
    if(currIndex){
        qsort(symbolArray,currIndex,sizeof(Symbol*),compareSymbolValue);
    }
    return currIndex;
}

