#include <SectHeader.h>
#include <Iterator.h>
#include <Function.h>
#include <LineInfoFinder.h>
#include <XCoffFile.h>
#include <BasicBlockCounter.h>

BasicBlockCounter::BasicBlockCounter(XCoffFile* xcoff,char* extension,uint32_t phaseNo)
    : CommonMethods(xcoff,extension,phaseNo,1,NULL)
{
}

void BasicBlockCounter::selectInstrumentationPoints(char* inpFile){
    
    terminationBlock = findTerminationBlock();
    ASSERT(terminationBlock && "FATAL : There is no exit block");

    uint32_t allBlocksCount = xcoffFile->getNumberOfBlocks();

    bool insertAppExitBlock = true;

    if (allBlocksCount){
        blocksToInstrument = new BasicBlock*[allBlocksCount];
        xcoffFile->getAllBlocks(blocksToInstrument);

        for (uint32_t i = 0; i < allBlocksCount; i++){
            BasicBlock* bb = blocksToInstrument[i];
            if (filterBlockForInst(bb)){
                PRINT_DEBUG("Filtering basic block %#12llx",bb->getHashCode().getValue());
                numOfBlocksFiltered++;
                continue;
            }
            if (insertAppExitBlock && (bb == terminationBlock)){
                PRINT_INFOR("The exit block sequence id will be %#x",numberOfInstPoints);
                insertAppExitBlock = false;
            }
            blocksToInstrument[numberOfInstPoints++] = bb;
        }
    } else {
        PRINT_ERROR("There is no basic block identified in the executable yet, is executable parsed");
    }

    if (!numberOfInstPoints){
        PRINT_ERROR("There is no basic block passed the filter");
    }

    if (insertAppExitBlock){
        blocksToInstrument[numberOfInstPoints++] = terminationBlock;
    }

    PRINT_INFOR("**** Number of basic blocks considered in instrumentation %d ******",numberOfInstPoints);
}

uint32_t BasicBlockCounter::byteCountForInst(uint32_t instPointIdx,uint64_t instStubAddress,TextSection* textSect){
    BasicBlock* block = blocksToInstrument[instPointIdx];

    uint64_t instPointAddress = 0;

    if(!block->findFirstInstPoint(&instPointAddress))
        return 0;

    ASSERT(textSect->inRange(instPointAddress));

    if(!Instruction::isInJumpInsnRange(instPointAddress,instStubAddress))
        return 0;

    int32_t blockCounterOffset = getDataOffsetForInst(instPointIdx);
    bool isApplicationExit = (block == terminationBlock);

    uint32_t minInsnCount = 8;
    uint32_t tocUpdateInsnCount = 0;
    uint32_t indexAccessInsnCount = 0;
    uint32_t exitInsnCount = 0;

    if(tocXorUpdateVal){
        tocUpdateInsnCount += 2;
        if(SECOND_HALFWORD(tocXorUpdateVal)){
            tocUpdateInsnCount += 2;
        }
    }
    if(!Instruction::isInLoadOffsetInsnRange(blockCounterOffset)){
        indexAccessInsnCount += 4;
    }
    if(isApplicationExit){
        tocUpdateInsnCount *= 2;
        exitInsnCount = 11;
    }
    uint32_t insnCount = minInsnCount + tocUpdateInsnCount + 
                         indexAccessInsnCount + exitInsnCount;

    return (insnCount*sizeof(uint32_t));
}

uint32_t BasicBlockCounter::generateCodeForInst(uint32_t instPointIdx,
                                                    uint64_t instStubAddress,TextSection* textSect,
                                                    BaseGen* gen,uint32_t genBufferOffset)
{
    ASSERT(gen->getParsedBase() == (Base*)textSect);
    BasicBlock* block = blocksToInstrument[instPointIdx];
    ASSERT(block);
    ASSERT(extendedDataSize && dataBufferEntryCount);

    uint64_t instPointAddress = 0;

    if(!block->findFirstInstPoint(&instPointAddress)){
        PRINT_INFOR("Basic Block %#18llx does not have suitable location for inst",
                    block->getHashCode().getValue());
        return 0;
    }
    ASSERT(textSect->inRange(instPointAddress));

    if(!Instruction::isInJumpInsnRange(instPointAddress,instStubAddress)){
        PRINT_INFOR("Basic Block %#18llx can not be instrumented due to branch offset limit",
                    block->getHashCode().getValue());
        return 0;
    }

    AddressIterator ait = block->getInstructionIterator();
    ait.skipTo(instPointAddress);
    Instruction origInsn = textSect->readInstruction(&ait);

    bool isApplicationExit = (block == terminationBlock);

    uint32_t savedReg = 31;
    uint32_t scratchReg = 30;
    ASSERT(savedReg > LAST_VOLATILE_REG);
    ASSERT(scratchReg > LAST_VOLATILE_REG);

    int32_t savedRegSaveOffset   = getGPRSaveOffset(savedReg);
    int32_t scratchRegSaveOffset = getGPRSaveOffset(scratchReg);
    int32_t blockCounterOffset   = getDataOffsetForInst(instPointIdx);

    uint32_t maxInsnCount = (maxByteCountPerInst() / sizeof(uint32_t));

    uint32_t insnCount = 0;
    Instruction* insnBuffer = new Instruction[maxInsnCount];

    insnBuffer[insnCount++] = origInsn;

    updateTOCRegister(insnBuffer,&insnCount);

    insnBuffer[insnCount++] = Instruction::generateStoreDouble(savedReg,REG_TOC,savedRegSaveOffset);

    bool oneWordCounter = (sizeof(uint32_t) == bytesPerBufferEntry());

    if(!Instruction::isInLoadOffsetInsnRange(blockCounterOffset)){
        insnBuffer[insnCount++] = Instruction::generateStoreDouble(scratchReg,REG_TOC,scratchRegSaveOffset);
        insnBuffer[insnCount++] = Instruction::generateLoad32BitHigh(scratchReg,blockCounterOffset);
        insnBuffer[insnCount++] = Instruction::generateLoad32BitLow(scratchReg,blockCounterOffset);
        if(oneWordCounter){
            insnBuffer[insnCount++] = Instruction::generateLoadWordIndx(savedReg,REG_TOC,scratchReg);
        } else {
            insnBuffer[insnCount++] = Instruction::generateLoadDoubleIndx(savedReg,REG_TOC,scratchReg);
        }

        insnBuffer[insnCount++] = Instruction::generateIncrement(savedReg,1);
        if(oneWordCounter){
            insnBuffer[insnCount++] = Instruction::generateStoreWordIndx(savedReg,REG_TOC,scratchReg);
        } else {
            insnBuffer[insnCount++] = Instruction::generateStoreDoubleIndx(savedReg,REG_TOC,scratchReg);
        }
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,REG_TOC,scratchRegSaveOffset);
    } else {
        if(oneWordCounter){
            insnBuffer[insnCount++] = Instruction::generateLoadWord(savedReg,REG_TOC,blockCounterOffset);
        } else {
            insnBuffer[insnCount++] = Instruction::generateLoadDouble(savedReg,REG_TOC,blockCounterOffset);
        }
        insnBuffer[insnCount++] = Instruction::generateIncrement(savedReg,1);
        if(oneWordCounter){
            insnBuffer[insnCount++] = Instruction::generateStoreWord(savedReg,REG_TOC,blockCounterOffset);
        } else {
            insnBuffer[insnCount++] = Instruction::generateStoreDouble(savedReg,REG_TOC,blockCounterOffset);
        }
    }

    if(isApplicationExit){
        insnBuffer[insnCount++] = Instruction::generateStoreDouble(scratchReg,REG_TOC,scratchRegSaveOffset);

        generateFuncCallFull(insnBuffer,&insnCount,savedReg,instStubAddress,libraryFuncStubAddrs[0]);

        insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,REG_TOC,scratchRegSaveOffset);
    }

    insnBuffer[insnCount++] = Instruction::generateLoadDouble(savedReg,REG_TOC,savedRegSaveOffset);

    updateTOCRegister(insnBuffer,&insnCount);

    uint64_t fromAddr = instStubAddress + (insnCount * sizeof(uint32_t));
    uint64_t toAddr   = instPointAddress + sizeof(uint32_t);
    insnBuffer[insnCount++] = Instruction::generateJumpInsn(fromAddr,toAddr);
    insnBuffer[insnCount++] = Instruction(instPointIdx);

    ASSERT(insnCount <= maxInsnCount);

    gen->writeInstructions(genBufferOffset,insnCount,insnBuffer);

    delete[] insnBuffer;

    fromAddr = instPointAddress;
    toAddr   = instStubAddress;
    Instruction origJumpToStub = Instruction::generateJumpInsn(fromAddr,toAddr);
    gen->writeWord(instPointAddress-textSect->getSectHeader()->GET(s_vaddr),origJumpToStub.bits());

    return (insnCount*sizeof(uint32_t));
}

bool BasicBlockCounter::filterBlockForInst(BasicBlock* bb){
    if(!bb)
        return true;
    if(bb->isTrace() || bb->isJumpTable())
        return true;
    if((bb->getNumOfMemoryOps() == 0) &&
       (bb->getNumOfFloatPOps() == 0))
       return true;
    return false;
}

void BasicBlockCounter::getSharedLibraryPathAndObj(char** path,char** name,char** obj){
    *path = getPathToInstLib();
    *obj  = "";
    if(xcoffFile->is64Bit()){
        *name = "libcntrFunc_64.a";
    } else {
        *name = "libcntrFunc.a";
    }
    ASSERT(path && name && obj);
}

void BasicBlockCounter::initializeReservedData(DataSection* dataSect,BaseGen* gen){
    gen->writeString(commentAddress-dataSect->getSectHeader()->GET(s_vaddr),
                     sharedLibraryInfoStr,strlen(sharedLibraryInfoStr) + 1);
}
