#include <SectHeader.h>
#include <Iterator.h>
#include <Function.h>
#include <LineInfoFinder.h>
#include <XCoffFile.h>
#include <CountAllBlocks.h>
#include <fstream>
using namespace std;

char* CountAllBlocks::staticPrintExtension = "static";

CountAllBlocks::CountAllBlocks(XCoffFile* xcoff,char* extension,uint32_t phaseNo)
    : XCoffFileGen(xcoff,extension,phaseNo,NULL),libraryFunctionCount(1),
      blocksToInstrument(NULL),terminationBlock(NULL),numOfBlocksFiltered(0),
      extendedBeginAddr(0),
      execTOCAddress(0),tocXorUpdateVal(0),tocDistUpdateVal(0),
      sharedLibraryInfoStr(NULL)
{
}

void CountAllBlocks::selectInstrumentationPoints(char* inpFile){
    
    terminationBlock = findTerminationBlock();
    ASSERT(terminationBlock && "FATAL : There is no exit block");

    uint32_t allBlocksCount = xcoffFile->getNumberOfBlocks();

    bool insertAppExitBlock = true;

    if (allBlocksCount){
        blocksToInstrument = new BasicBlock*[allBlocksCount];
        xcoffFile->getAllBlocks(blocksToInstrument);

        for (uint32_t i = 0; i < allBlocksCount; i++){
            BasicBlock* bb = blocksToInstrument[i];
            if (filterBlockForInst(bb)){
                PRINT_DEBUG("Filtering basic block %#12llx",bb->getHashCode().getValue());
                numOfBlocksFiltered++;
                continue;
            }
            if (insertAppExitBlock && (bb == terminationBlock)){
                PRINT_INFOR("The exit block sequence id will be %#x",numberOfInstPoints);
                insertAppExitBlock = false;
            }
            blocksToInstrument[numberOfInstPoints++] = bb;
        }
    } else {
        PRINT_ERROR("There is no basic block identified in the executable yet, is executable parsed?");
    }

    if (!numberOfInstPoints){
        PRINT_ERROR("There is no basic block passed the filter");
    }

    if (insertAppExitBlock){
        blocksToInstrument[numberOfInstPoints++] = terminationBlock;
    }

    PRINT_INFOR("**** Number of basic blocks considered in instrumentation %d ******",numberOfInstPoints);
}

uint32_t CountAllBlocks::byteCountForInst(uint32_t instPointIdx,uint64_t instStubAddress,TextSection* textSect){
    BasicBlock* block = blocksToInstrument[instPointIdx];

    uint64_t instPointAddress = 0;

    if(!block->findFirstInstPoint(&instPointAddress)){
        return 0;
    }

    ASSERT(textSect->inRange(instPointAddress));

    if(!Instruction::isInJumpInsnRange(instPointAddress,instStubAddress))
        return 0;

    int32_t blockCounterOffset = getDataOffsetForInst(instPointIdx);
    bool isApplicationExit = (block == terminationBlock);

    uint32_t minInsnCount = 8;
    uint32_t tocUpdateInsnCount = 0;
    uint32_t indexAccessInsnCount = 0;
    uint32_t exitInsnCount = 0;

    if(tocXorUpdateVal){
        tocUpdateInsnCount += 2;
        if(SECOND_HALFWORD(tocXorUpdateVal)){
            tocUpdateInsnCount += 2;
        }
    }
    if(!Instruction::isInLoadOffsetInsnRange(blockCounterOffset)){
        indexAccessInsnCount += 4;
    }
    if(isApplicationExit){
        tocUpdateInsnCount *= 2;
        exitInsnCount = 11;
    }
    uint32_t insnCount = minInsnCount + tocUpdateInsnCount + 
                         indexAccessInsnCount + exitInsnCount;

    return (insnCount*sizeof(uint32_t));
}

uint32_t CountAllBlocks::generateCodeForInst(uint32_t instPointIdx,
                                                    uint64_t instStubAddress,TextSection* textSect,
                                                    BaseGen* gen,uint32_t genBufferOffset)
{
    ASSERT(gen->getParsedBase() == (Base*)textSect);
    BasicBlock* block = blocksToInstrument[instPointIdx];
    ASSERT(block);
    ASSERT(extendedDataSize && dataBufferEntryCount);

    uint64_t instPointAddress = 0;

    if(!block->findFirstInstPoint(&instPointAddress)){
        PRINT_INFOR("Basic Block %#18llx does not have suitable location for inst",
                    block->getHashCode().getValue());
        return 0;
    }
    ASSERT(textSect->inRange(instPointAddress));

    if(!Instruction::isInJumpInsnRange(instPointAddress,instStubAddress)){
        PRINT_INFOR("Basic Block %#18llx can not be instrumented due to branch offset limit",
                    block->getHashCode().getValue());
        return 0;
    }

    AddressIterator ait = block->getInstructionIterator();
    ait.skipTo(instPointAddress);
    Instruction origInsn = textSect->readInstruction(&ait);

    bool isApplicationExit = (block == terminationBlock);

    uint32_t savedReg = 31;
    uint32_t scratchReg = 30;
    ASSERT(savedReg > LAST_VOLATILE_REG);
    ASSERT(scratchReg > LAST_VOLATILE_REG);

    int32_t savedRegSaveOffset   = getGPRSaveOffset(savedReg);
    int32_t scratchRegSaveOffset = getGPRSaveOffset(scratchReg);
    int32_t blockCounterOffset   = getDataOffsetForInst(instPointIdx);

    uint32_t maxInsnCount = (maxByteCountPerInst() / sizeof(uint32_t));

    uint32_t insnCount = 0;
    Instruction* insnBuffer = new Instruction[maxInsnCount];

/* execute the instruction from the original code that has been replaced */
    insnBuffer[insnCount++] = origInsn;

    updateTOCRegister(insnBuffer,&insnCount);

/* store savedReg to data section (save savedReg) */
    insnBuffer[insnCount++] = Instruction::generateStoreDouble(savedReg,REG_TOC,savedRegSaveOffset);

    bool oneWordCounter = (sizeof(uint32_t) == bytesPerBufferEntry());

/* if blockCounterOffset is too large to be used as an immediate in a load, special steps must be taken */
    if(!Instruction::isInLoadOffsetInsnRange(blockCounterOffset)){

    /* store scratchReg to data section (save scratchReg) */
        insnBuffer[insnCount++] = Instruction::generateStoreDouble(scratchReg,REG_TOC,scratchRegSaveOffset);
    /* load counter offset in data section to scratchReg */
        insnBuffer[insnCount++] = Instruction::generateLoad32BitHigh(scratchReg,blockCounterOffset);
        insnBuffer[insnCount++] = Instruction::generateLoad32BitLow(scratchReg,blockCounterOffset);

    /* load the counter from ext. data section to savedReg */
        if(oneWordCounter){
            insnBuffer[insnCount++] = Instruction::generateLoadWordIndx(savedReg,REG_TOC,scratchReg);
        } else {
            insnBuffer[insnCount++] = Instruction::generateLoadDoubleIndx(savedReg,REG_TOC,scratchReg);
        }

    /* increment counter */
        insnBuffer[insnCount++] = Instruction::generateIncrement(savedReg,1);

    /* store the counter from savedReg to data section */
        if(oneWordCounter){
            insnBuffer[insnCount++] = Instruction::generateStoreWordIndx(savedReg,REG_TOC,scratchReg);
        } else {
            insnBuffer[insnCount++] = Instruction::generateStoreDoubleIndx(savedReg,REG_TOC,scratchReg);
        }
    /* load scrtachReg from data section (restore scratchReg) */
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,REG_TOC,scratchRegSaveOffset);

/* if blockCounterOffset is small enough to be used as the immediate in a load */
    } else {
    /* load the counter from ext. data section to savedReg */
        if(oneWordCounter){
            insnBuffer[insnCount++] = Instruction::generateLoadWord(savedReg,REG_TOC,blockCounterOffset);
        } else {
            insnBuffer[insnCount++] = Instruction::generateLoadDouble(savedReg,REG_TOC,blockCounterOffset);
        }

    /* increment counter */
        insnBuffer[insnCount++] = Instruction::generateIncrement(savedReg,1);

    /* store the counter from savedReg to data section */
        if(oneWordCounter){
            insnBuffer[insnCount++] = Instruction::generateStoreWord(savedReg,REG_TOC,blockCounterOffset);
        } else {
            insnBuffer[insnCount++] = Instruction::generateStoreDouble(savedReg,REG_TOC,blockCounterOffset);
        }
    }


/* a function must be called at the exit point of the program to print all of the information collected during
   the program run. */
    if(isApplicationExit){

/* save scratch reg, generate a function call, restore scratch reg */
        insnBuffer[insnCount++] = Instruction::generateStoreDouble(scratchReg,REG_TOC,scratchRegSaveOffset);
        generateFuncCallFull(insnBuffer,&insnCount,savedReg,instStubAddress,libraryFuncStubAddrs[0]);
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,REG_TOC,scratchRegSaveOffset);
    }

/* load savedReg from data section (restore savedReg) */
    insnBuffer[insnCount++] = Instruction::generateLoadDouble(savedReg,REG_TOC,savedRegSaveOffset);

    updateTOCRegister(insnBuffer,&insnCount);

    uint64_t fromAddr = instStubAddress + (insnCount * sizeof(uint32_t));
    uint64_t toAddr   = instPointAddress + sizeof(uint32_t);

/* jump from instrumentation code to original code */
    insnBuffer[insnCount++] = Instruction::generateJumpInsn(fromAddr,toAddr);
/* put a "tag" instruction (will NOT be execited) after the instrumentation sequence. This makes it easier to
   find instrumentation code for debugging and such */
    insnBuffer[insnCount++] = Instruction(instPointIdx);

    ASSERT(insnCount <= maxInsnCount);

    gen->writeInstructions(genBufferOffset,insnCount,insnBuffer);

    delete[] insnBuffer;

    fromAddr = instPointAddress;
    toAddr   = instStubAddress;
    Instruction origJumpToStub = Instruction::generateJumpInsn(fromAddr,toAddr);
    gen->writeWord(instPointAddress-textSect->getSectHeader()->GET(s_vaddr),origJumpToStub.bits());

    return (insnCount*sizeof(uint32_t));
}

/* returns true if the basic block is code as opposed to a "false block" found in a jump table or trace */
bool CountAllBlocks::filterBlockForInst(BasicBlock* bb){
    if(!bb)
        return true;
    if(bb->isTrace() || bb->isJumpTable())
        return true;
    return false;
}

void CountAllBlocks::getSharedLibraryPathAndObj(char** path,char** name,char** obj){

    *path = getPathToInstLib();
    *obj  = "";
    if(xcoffFile->is64Bit()){
        *name = "libcntrFunc_64.a";
    } else {
        *name = "libcntrFunc.a";
    }
    ASSERT(path && name && obj);
}

void CountAllBlocks::reserveDataForInstrumentation(){

    extendedDataSize = 0;
    extendedBeginAddr = 0;
    dataBufferEntryCount = 0;

    uint64_t dataSectionBeginAddr = xcoffFile->getDataSectionVAddr();
    uint32_t dataSectionSize = xcoffFile->getDataSectionSize();

    execTOCAddress = xcoffFile->getTOCAddress();

    uint64_t currentFreeAddress = dataSectionBeginAddr + dataSectionSize;

    uint32_t incrementSize = 0;
    if (xcoffFile->is64Bit()){
        currentFreeAddress = nextAlignAddressDouble(currentFreeAddress);
        incrementSize = sizeof(uint64_t);
    } else {
        currentFreeAddress = nextAlignAddressWord(currentFreeAddress);
        incrementSize = sizeof(uint32_t);
    }

    extendedBeginAddr = currentFreeAddress;

/* reserve space for function addrs from the shared library */
    for (uint32_t i = 0; i < getNumOfSharedLibFuncs(); i++){
        libraryFunctionStubAddrLocs[i] = currentFreeAddress;
        currentFreeAddress += incrementSize;
    }


/* reserve space for registers to be saved */
    currentFreeAddress = nextAlignAddressWord(currentFreeAddress);

    saveAddressCondReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint32_t);
    saveAddressFPCondReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint32_t);

    currentFreeAddress = nextAlignAddressDouble(currentFreeAddress);

    saveAddressLinkReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint64_t);
    saveAddressCtrReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint64_t);
    saveAddressExpReg = currentFreeAddress;
    currentFreeAddress += sizeof(uint64_t);

    saveAddressGprRegs = currentFreeAddress;
    currentFreeAddress += (sizeof(uint64_t) * NUM_OF_GPR_REGS);
    saveAddressFprRegs = currentFreeAddress;
    currentFreeAddress += (sizeof(uint64_t) * NUM_OF_FPR_REGS);
        

/* reserve space for a "comment", which is just some information about the program. first the 
   comment string is created, then we figure out its size */
    char* ptr = xcoffFile->getXCoffFileName();
    char* sepPtr = strrchr(ptr,'/');
    if(sepPtr){
        ptr = (sepPtr + 1);
    }
    char tmpStr[__MAX_STRING_SIZE];
    if(!phaseIndex){
        sprintf(tmpStr,"%s 0 %s %d",ptr,getFileExtension(),
                getNumberOfInstPoints(NULL));
    } else {
        sprintf(tmpStr,"%s %d %s %d",ptr,getPhaseIndex(),getFileExtension(),
                getNumberOfInstPoints(NULL));
    }
    commentAddress = currentFreeAddress;
    sharedLibraryInfoStr = strdup(tmpStr);
    currentFreeAddress += (strlen(sharedLibraryInfoStr) + 1);

/* reserve space for the block counter */
    dataBufferEntryCount = howManyBufferEntry();
    incrementSize = bytesPerBufferEntry();
    if(incrementSize == sizeof(uint32_t)){
        currentFreeAddress = nextAlignAddressWord(currentFreeAddress);
    } else {
        currentFreeAddress = nextAlignAddressDouble(currentFreeAddress);
    }
    dataBufferAddress = currentFreeAddress;
    currentFreeAddress += (dataBufferEntryCount * incrementSize);


    currentFreeAddress = nextAlignAddressDouble(currentFreeAddress);

    extendedDataSize = currentFreeAddress - (dataSectionBeginAddr + dataSectionSize);

/* If the end of the updated data section has an address that is too large to be used as a load immediate value */
    tocXorUpdateVal = 0;
    if(!Instruction::isInLoadOffsetInsnRange((int32_t)(dataBufferAddress-execTOCAddress))){
        tocXorUpdateVal = (uint32_t)(extendedBeginAddr ^ execTOCAddress);
        tocDistUpdateVal = (uint32_t)(extendedBeginAddr - execTOCAddress);
    }

    PRINT_INFOR("**** Additional data section inc  is %d ***",extendedDataSize);
    PRINT_INFOR("**** Number of entries to process is %18d ***",dataBufferEntryCount);
    PRINT_INFOR("**** Start address for toc        is %#18llx (%lld) ***",
        execTOCAddress,execTOCAddress-dataSectionBeginAddr);
    PRINT_INFOR("**** Start address for additional is %#18llx (%lld) ***",
        extendedBeginAddr,extendedBeginAddr-execTOCAddress);
    PRINT_INFOR("**** The tocXorUpdateVal for save is %x %d ***",
        tocXorUpdateVal,tocDistUpdateVal);
    PRINT_INFOR("**** Start address for the gprs   is %#18llx (%lld) ***",
        saveAddressGprRegs,saveAddressGprRegs-execTOCAddress);
    PRINT_INFOR("**** Start address for the comm   is %#18llx (%lld) ***",
        commentAddress,commentAddress-execTOCAddress);
    PRINT_INFOR("**** Start address for the buffer is %#18llx (%lld) ***",
        dataBufferAddress,dataBufferAddress-execTOCAddress);
    PRINT_INFOR("**** End address for the buffer   is %#18llx (%lld) ***",
        currentFreeAddress,currentFreeAddress-execTOCAddress);
}

void CountAllBlocks::initializeReservedData(DataSection* dataSect,BaseGen* gen){
/* write the comment to its reserved space */
    gen->writeString(commentAddress-dataSect->getSectHeader()->GET(s_vaddr),
                     sharedLibraryInfoStr,strlen(sharedLibraryInfoStr) + 1);
}

BasicBlock* CountAllBlocks::findTerminationBlock(){

    uint32_t allBlocksCount = xcoffFile->getNumberOfBlocks();
    if(!allBlocksCount)
        return NULL;

    BasicBlock** allBlocks = new BasicBlock*[allBlocksCount];
    xcoffFile->getAllBlocks(allBlocks);

    BasicBlock* mpiexit = NULL;
    BasicBlock* regexit = NULL;
    BasicBlock* ddiexit = NULL;

    for (uint32_t i = 0; i < allBlocksCount; i++){
        BasicBlock* bb = allBlocks[i];
        if (bb->isTrace() || bb->isJumpTable() || bb->getIndex())
            continue;
        Function* owner = bb->getFlowGraph()->getFunction();
        if (!ddiexit && owner->isAnySymbolA(".DDI_Finalize")){
            ddiexit = bb;
        } 
        if (!mpiexit && owner->isAnySymbolA(".mpi_finalize")){
            mpiexit = bb;
        } 
        if (!regexit && owner->isAnySymbolA(".exit")){
            regexit = bb;
        }
    }
    delete[] allBlocks;

    if (ddiexit) {
        PRINT_INFOR("The exit function .DDI_Finalize is taken");
        return ddiexit;
    }
    if (mpiexit){
    PRINT_INFOR("The exit function .mpi_finalize is taken");
        return mpiexit;
    }
    if (regexit){
        PRINT_INFOR("The exit function .exit is taken");
        return regexit;
    }
    return NULL;
}

void CountAllBlocks::printInstrumentationPoints(){

    char instBlockFileName[__MAX_STRING_SIZE];
    if (!getPhaseIndex()){
        sprintf(instBlockFileName,"%s.%s.%s",xcoffFile->getXCoffFileName(),getFileExtension(),staticPrintExtension);
    } else {
        sprintf(instBlockFileName,"%s.phase.%d.%s.%s",xcoffFile->getXCoffFileName(),getPhaseIndex(),
                getFileExtension(),staticPrintExtension);
    }

    ofstream instBlockFile;
    instBlockFile.open(instBlockFileName);
    if (!instBlockFile){
         PRINT_ERROR("Instrumented block infor can not be opened [%s]",instBlockFileName);
    }

    uint32_t memOpCount = 0;
    uint32_t fpOpCount = 0;
    uint32_t insnCount = 0;

    for (uint32_t i=0;i<numberOfInstPoints;i++){
        memOpCount += blocksToInstrument[i]->getNumOfMemoryOps();
        fpOpCount += blocksToInstrument[i]->getNumOfFloatPOps();
        insnCount += blocksToInstrument[i]->getInstructionCount();
    }

    TextSection* lineInfoSect = (TextSection*)xcoffFile->getRawSection(xcoffFile->getTextSectionIndex());
    LineInfoFinder* lineInfoFinder = lineInfoSect->getLineInfoFinder();

    instBlockFile << "# appname   = " << xcoffFile->getXCoffFileName() << endl;
    instBlockFile << "# appsize   = " << xcoffFile->getFileSize() << endl;
    instBlockFile << "# extension = " << getFileExtension() << endl;
    instBlockFile << "# phase     = " << getPhaseIndex() << endl;
    instBlockFile << "# type      = " << getGenClassName() << endl;
    instBlockFile << "# candidate = " << xcoffFile->getNumberOfBlocks() << endl;
    instBlockFile << "# blocks    = " << numberOfInstPoints << endl;
    instBlockFile << "# memops    = " << memOpCount << endl;
    instBlockFile << "# fpops     = " << fpOpCount << endl;
    instBlockFile << "# insns     = " << insnCount << endl;
    instBlockFile << "# buffer    = " << howManyBufferEntry() << endl;
    instBlockFile << "# <sequence> <block_uid> <memop> <fpop> <insn> <line> <fname> <hex_uid> <vaddr>" << endl;

    for (uint32_t i = 0; i < numberOfInstPoints; i++){
        uint64_t unq = blocksToInstrument[i]->getHashCode().getValue();
        Function* owner = blocksToInstrument[i]->getFlowGraph()->getFunction();
        char names[__MAX_STRING_SIZE];
        names[0] = '\0';
        owner->getAllSymbolNames(names,__MAX_STRING_SIZE,true);
        if(!strlen(names)){
            strcpy(names,"__name__unknown__");
        }
        char* absFile = NULL;
        uint32_t absLine  =0;
        if(lineInfoFinder){
            uint64_t addr = blocksToInstrument[i]->getBaseAddress();
            absFile = lineInfoFinder->getFileName(addr);
            absLine = lineInfoFinder->getLineNumberInFile(addr);
        }
        instBlockFile << i << "\t" << unq 
                      << "\t" << blocksToInstrument[i]->getNumOfMemoryOps()
                      << "\t" << blocksToInstrument[i]->getNumOfFloatPOps()
                      << "\t" << blocksToInstrument[i]->getInstructionCount()
                      << "\t" << (absFile ? absFile : "__FILE_UNK__") << ":" << absLine 
                      << "\t" << names
                      << "\t# 0x" << hex << unq << "\t0x" << blocksToInstrument[i]->getBaseAddress() << dec 
                      << endl;    
    }
    instBlockFile.close();
}


void CountAllBlocks::updateTOCRegister(Instruction* insnBuffer,uint32_t* insnIndex){
    uint16_t tocUpdateValLow = FIRST_HALFWORD(tocXorUpdateVal);
    uint16_t tocUpdateValHigh = SECOND_HALFWORD(tocXorUpdateVal);
    if(tocXorUpdateVal){
        if(tocUpdateValHigh){
            insnBuffer[(*insnIndex)++] = Instruction::generateXorImmShifted(REG_TOC,REG_TOC,tocUpdateValHigh);
        }
        insnBuffer[(*insnIndex)++] = Instruction::generateXorImm(REG_TOC,REG_TOC,tocUpdateValLow);
    }
}

void CountAllBlocks::generateFuncCallFull(Instruction* insnBuffer,uint32_t* insnIndex,
                                                uint32_t reg,uint64_t stubAddr,uint64_t funcAddr)
{
//    PRINT_INFOR("Generating a function call with args %d %d %#llx %#llx", *insnIndex, reg, stubAddr, funcAddr);

/* save CR register */
    insnBuffer[(*insnIndex)++] = Instruction::generateMoveFromCR(reg);
    insnBuffer[(*insnIndex)++] = Instruction::generateStoreWord(reg,REG_TOC,getCRSaveOffset());

/* save LNK register */
    insnBuffer[(*insnIndex)++] = Instruction::generateMoveFromSPR(reg,REG_LNK);
    insnBuffer[(*insnIndex)++] = Instruction::generateStoreDouble(reg,REG_TOC,getLNKSaveOffset());

    updateTOCRegister(insnBuffer,insnIndex);

/* calculate target and source addresses for function call */
    uint64_t fromAddr = stubAddr + ((*insnIndex) * sizeof(uint32_t));
    uint64_t toAddr   = funcAddr;

/* generate call to function */
    insnBuffer[(*insnIndex)++] = Instruction::generateCallToImmediate(fromAddr,toAddr);


    updateTOCRegister(insnBuffer,insnIndex);

/* restore original LNK register value */
    insnBuffer[(*insnIndex)++] = Instruction::generateLoadDouble(reg,REG_TOC,getLNKSaveOffset());
    insnBuffer[(*insnIndex)++] = Instruction::generateMoveToSPR(reg,REG_LNK);

/* restore original CR register value */
    insnBuffer[(*insnIndex)++] = Instruction::generateLoadWord(reg,REG_TOC,getCRSaveOffset());
    insnBuffer[(*insnIndex)++] = Instruction::generateMoveToCR(reg);
}

uint32_t CountAllBlocks::generateSharedLibFuncWrapper(
    uint32_t libFuncIdx,uint64_t funcCallAddr,uint32_t genBufferOffset,BaseGen* gen)
{


    uint32_t insnCount = 0;
    Instruction* insnBuffer = new Instruction[byteCountForSharedLibFuncWrappers() / sizeof(uint32_t)];

    uint32_t scratchReg = 6;
    uint32_t fpScratchReg = 3;
    uint32_t savedReg = 31;
    uint32_t tocHoldReg = 30;
    uint32_t firstArgReg = 3;
    uint32_t secondArgReg = 4;
    uint32_t thirdArgReg = 5;

    ASSERT(scratchReg == 6);
    ASSERT(fpScratchReg == 3);
    ASSERT(savedReg == 31);
    ASSERT(tocHoldReg == 30);
    ASSERT(firstArgReg == 3);
    ASSERT(secondArgReg == 4);
    ASSERT(thirdArgReg == 5);

    int32_t saveOffset = 0x0;

    updateTOCRegister(insnBuffer,&insnCount);

/* store all GP registers to extended data section */
    for(uint32_t i=0;i<=LAST_VOLATILE_REG;i++){
        saveOffset = getGPRSaveOffset(i);
        insnBuffer[insnCount++] = Instruction::generateStoreDouble(i,REG_TOC,saveOffset);
    }

/* store all FP registers to extended data section */
    for(uint32_t i=0;i<=LAST_VOLATILE_REG;i++){
        saveOffset = getFPRSaveOffset(i);
        insnBuffer[insnCount++] = Instruction::generateStoreDoubleFloat(i,REG_TOC,saveOffset);
    }

/* save CTR Register */
    insnBuffer[insnCount++] = Instruction::generateMoveFromSPR(scratchReg,REG_CTR);
    insnBuffer[insnCount++] = Instruction::generateStoreDouble(scratchReg,REG_TOC,getCTRSaveOffset());

/* save EXR Register */
    insnBuffer[insnCount++] = Instruction::generateMoveFromSPR(scratchReg,REG_XER);
    insnBuffer[insnCount++] = Instruction::generateStoreDouble(scratchReg,REG_TOC,getEXRSaveOffset());

/* save FPSCR Register */
    insnBuffer[insnCount++] = Instruction::generateMoveFromFPSCR(scratchReg);
    insnBuffer[insnCount++] = Instruction::generateStoreWordFloat(scratchReg,REG_TOC,getFPSCRSaveOffset());

/* put buffer address (in extended data section) into first argument register to be passed to function */
    uint32_t bufferOffset = getDataOffsetForInst(0);
    insnBuffer[insnCount++] = Instruction::generateMoveReg(REG_TOC,firstArgReg);
    insnBuffer[insnCount++] = Instruction::generateIncrement(firstArgReg,bufferOffset);

/* put the number of buffer entries into second argument register to be passed to function */
    uint32_t bufferCount = getDataBufferEntryCount();
    insnBuffer[insnCount++] = Instruction::generateLoad32BitHigh(secondArgReg,bufferCount);
    insnBuffer[insnCount++] = Instruction::generateLoad32BitLow(secondArgReg,bufferCount);

/* put comment address (in extended data section) into third argument register to be passed to function */
    uint32_t commentOffset = getCommentAddressOffset();
    insnBuffer[insnCount++] = Instruction::generateMoveReg(REG_TOC,thirdArgReg);
    insnBuffer[insnCount++] = Instruction::generateIncrement(thirdArgReg,commentOffset);

/* save LNK reg to savedReg */
    insnBuffer[insnCount++] = Instruction::generateMoveFromSPR(savedReg,REG_LNK);

/* save TOC reg to tocHoldReg */
    insnBuffer[insnCount++] = Instruction::generateMoveReg(REG_TOC,tocHoldReg);


    int32_t callOffset = funcCallAddr - execTOCAddress - tocDistUpdateVal;
    if(xcoffFile->is64Bit()){
/* move stack pointer (push a new stack frame) */
        insnBuffer[insnCount++] = Instruction::generateSPIncrementDouble(-SAFE_FRAME_LOC*sizeof(uint64_t));
/* load pointer to function address into scratchReg */
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,REG_TOC,callOffset);
/* Each function pointer is composed of the address of function instructions followed by the TOC register value 
   for the function. first we load the TOC register for the function into REG_TOC */
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(REG_TOC,scratchReg,sizeof(uint64_t));
/* Then we load the function address into scratchReg*/
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,scratchReg,0);
    } else {
        insnBuffer[insnCount++] = Instruction::generateSPIncrementWord(-SAFE_FRAME_LOC*sizeof(uint64_t));
        insnBuffer[insnCount++] = Instruction::generateLoadWord(scratchReg,REG_TOC,callOffset);
        insnBuffer[insnCount++] = Instruction::generateLoadWord(REG_TOC,scratchReg,sizeof(uint32_t));
        insnBuffer[insnCount++] = Instruction::generateLoadWord(scratchReg,scratchReg,0);
    }

/* move function address to CTR register */
    insnBuffer[insnCount++] = Instruction::generateMoveToSPR(scratchReg,REG_CTR);
/* call the function by calling the adddress in the CTR reg */
    insnBuffer[insnCount++] = Instruction::generateCallToCTR();

/* restore the original TOC register value */
    insnBuffer[insnCount++] = Instruction::generateMoveReg(tocHoldReg,REG_TOC);

/* restore the original LNK register value */
    insnBuffer[insnCount++] = Instruction::generateMoveToSPR(savedReg,REG_LNK);

/* restore original FPSCR register value */
    insnBuffer[insnCount++] = Instruction::generateLoadWordFloat(fpScratchReg,REG_TOC,getFPSCRSaveOffset());
    insnBuffer[insnCount++] = Instruction::generateMoveToFPSCR(fpScratchReg);

/* restore original EXR register value */
    insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,REG_TOC,getEXRSaveOffset());
    insnBuffer[insnCount++] = Instruction::generateMoveToSPR(scratchReg,REG_XER);

/* restore original CTR register value */
    insnBuffer[insnCount++] = Instruction::generateLoadDouble(scratchReg,REG_TOC,getCTRSaveOffset());
    insnBuffer[insnCount++] = Instruction::generateMoveToSPR(scratchReg,REG_CTR);

/* restore original FP register values */
    for(uint32_t i=0;i<=LAST_VOLATILE_REG;i++){
        saveOffset = getFPRSaveOffset(i);
        insnBuffer[insnCount++] = Instruction::generateLoadDoubleFloat(i,REG_TOC,saveOffset);
    }

/* restore original GP register values */
    for(uint32_t i=0;i<=LAST_VOLATILE_REG;i++){
        saveOffset = getGPRSaveOffset(i);
        insnBuffer[insnCount++] = Instruction::generateLoadDouble(i,REG_TOC,saveOffset);
    }

    updateTOCRegister(insnBuffer,&insnCount);

/* return to original code (address held in LNK reg) */
    insnBuffer[insnCount++] = Instruction::generateReturnToLnk();

    gen->writeInstructions(genBufferOffset,insnCount,insnBuffer);

    delete[] insnBuffer;

    return (insnCount*sizeof(uint32_t));
}
